import numpy as np
import orjson
from sqlitedict import SqliteDict
from cachetools import LRUCache, cached
from cachetools.keys import hashkey

POSITION_ATTRIBUTE = "rep_coord_nm"


class NoL2CacheException(Exception):
    pass


def cachekey(client):
    return hashkey(client.datastack_name)


@cached(LRUCache(16), key=cachekey)
def check_l2cache(client):
    return client.l2cache.has_cache()


def get_locs_remote(l2_ids, client):
    """Retrieve ids from the l2 cache service"""
    if check_l2cache(client) is False:
        raise NoL2CacheException(
            f"No L2 Cache service is running for datastack {client.datastack_name} and you cannot use PCGSkel without an L2 Cache. Contact your admin for more information."
        )
    l2_info = client.l2cache.get_l2data(list(l2_ids), attributes=[POSITION_ATTRIBUTE])
    l2loc = [l2_info[str(l2id)].get(POSITION_ATTRIBUTE, None) for l2id in l2_ids]

    l2means = np.array([x for x in l2loc if x is not None])
    is_cached = np.array([True if x is not None else False for x in l2loc])

    if len(l2means) == 0:
        l2means = np.zeros((0, 3), dtype=float)

    return l2means, is_cached


def get_locs_local(l2_ids, cache_file):
    l2means = []
    is_cached = []

    with SqliteDict(cache_file, decode=orjson.loads, flag="c") as cache_dict:
        for l2id in l2_ids:
            loc = cache_dict.get(str(l2id), None)
            if loc is not None:
                l2means.append(loc)
                is_cached.append(True)
            else:
                is_cached.append(False)

    if len(l2means) > 0:
        l2means = np.vstack(l2means)
    else:
        l2means = np.zeros((0, 3), dtype=float)

    return l2means, np.array(is_cached)


def lookup_cached_ids(l2_ids, remote_cache=False, cache_file=None, client=None):
    if remote_cache:
        return get_locs_remote(l2_ids, client=client)
    else:
        return get_locs_local(l2_ids, cache_file=cache_file)


def save_ids_to_cache(l2_ids, l2_locs, cache_file):
    """Save computed locations back to the l2 cache.

    Parameters
    ----------
    l2_ids : list-like
        List of N layer 2 ids
    l2_locs : np.array
        Nx3 array of locations
    cache_file : str

    """
    with SqliteDict(cache_file, encode=orjson.dumps, flag="c") as cache_dict:
        for k, v in zip(l2_ids, l2_locs):
            if not np.any(np.isnan(v)):
                cache_dict[str(k)] = v.tolist()
        cache_dict.commit()



import datetime
from typing import Optional

import numpy as np
import pandas as pd
from caveclient import CAVEclient
from meshparty import meshwork, skeleton, trimesh_io

from . import features

compartment_map = {
    "soma": 1,
    "axon": 2,
    "dendrite": 3,
}

def rebuild_meshwork(
    root_id: int,
    sk_verts: np.ndarray,
    sk_edges: np.ndarray,
    root: int,
    mesh_to_skel_map: np.ndarray,
    lvl2_ids: np.ndarray,
    radius: Optional[list] = None,
    compartments: Optional[list] = None,
    is_axon: bool = True,
    metadata: Optional[dict] = None,
    client: Optional[CAVEclient] = None,
    synapses: bool = False,
    timestamp: Optional[datetime.datetime] = None,
    restore_graph: bool = False,
    restore_properties: bool = False,
    synapse_reference_tables: dict = {},
):
    if metadata is None:
        metadata = {}
    skel = skeleton.Skeleton(
        vertices=sk_verts,
        edges=sk_edges,
        root=root,
        mesh_to_skel_map=mesh_to_skel_map,
        radius=radius,
        meta=metadata,
    )

    l2dict_mesh = {l2id: ii for ii, l2id in enumerate(lvl2_ids)}
    if restore_graph:
        gr = client.chunkedgraph.level2_chunk_graph(root_id)
        mesh_edges = np.apply_along_axis(lambda x: [l2dict_mesh[y] for y in x], 0, gr)

        locs = client.l2cache.get_l2data(lvl2_ids, attributes=["rep_coord_nm"])
        mesh_verts = [locs[str(l2id)].get("rep_coord_nm") for l2id in lvl2_ids]
    else:
        mesh_edges = []
        mesh_verts = sk_verts[mesh_to_skel_map]

    mesh = trimesh_io.Mesh(
        vertices=mesh_verts,
        faces=[[0, 0, 0]],  # Can error without faces,
        link_edges=mesh_edges,
    )

    nrn = meshwork.Meshwork(
        mesh=mesh,
        skeleton=skel,
        seg_id=root_id,
    )

    lvl2_df = pd.DataFrame({"lvl2_id": lvl2_ids, "mesh_ind": np.arange(len(lvl2_ids))})
    nrn.anno.add_annotations("lvl2_ids", lvl2_df, index_column="mesh_ind")

    if compartments is not None:
        compartment_df = pd.DataFrame(
            {"compartment": compartments[mesh_to_skel_map], "mesh_ind": np.arange(len(mesh_to_skel_map))}
        )
        nrn.anno.add_annotations("compartment", compartment_df, index_column="mesh_ind")
        if is_axon:
            if compartment_map['axon'] in compartments:
                is_axon = compartment_df.query(f'compartment == {compartment_map["axon"]}')['mesh_ind'].values
                nrn.anno.add_annotations("is_axon", is_axon, mask=True)
            else:
                nrn.anno.add_annotations("is_axon", [], mask=True)
    else:
        if is_axon:
            nrn.anno.add_annotations("is_axon", [], mask=True)

    if synapses:
        features.add_synapses(
            nrn,
            synapse_table=client.materialize.synapse_table,
            l2dict_mesh=l2dict_mesh,
            client=client,
            root_id=nrn.seg_id,
            pre=True,
            post=True,
            remove_self_synapse=True,
            timestamp=timestamp,
            live_query=True,
            metadata=False,
            synapse_point_resolution=[1,1,1], # Nanometers, same as skeleton.
            synapse_reference_tables=synapse_reference_tables,
        )

    if restore_properties:
        features.add_volumetric_properties(
            nrn,
            client=client,
        )
        features.add_segment_properties(nrn)

    return nrn


def get_meshwork_from_client(
    root_id: int,
    client: CAVEclient,
    synapses: bool = False,
    restore_graph: bool = False,
    restore_properties: bool = False,
    synapse_reference_tables: Optional[dict] = None,
    skeleton_version: Optional[int] = 4,
) -> meshwork.Meshwork:
    """Generate a meshwork file from the information on the skeleton service.
    Note: Requires skeleton service v0.12.3 or higher to be deployed.

    Parameters
    ----------
    root_id : int
        Object root id
    client : CAVEclient
        Initialized caveclient
    synapses : bool, optional
        Whether to add synapses under the annotations "pre_syn" and "post_syn", by default False
    restore_graph : bool, optional
        Whether to restore the level 2 graph, by default False.
        Adds a significant amount of time to the rehydration,
        however the `nrn.mesh` part will not be accurate if this is True.
    restore_properties : bool, optional
        Whether to restore `volume_properties` and `segment_properties`, by default False.
        Adds a significant amount of time to the rehydration.
    synapse_reference_tables : dict, optional
        A dictionary of synapse reference tables to attach to synapses.
        Keys are 'pre' and 'post', values are table names. Defaults to {}.
    skeleton_version : Optional[int], optional
        Which skeleton version to use, by default 4, which is the minimum needed for rehydration to work.

    Returns
    -------
    meshparty.meshwork.MeshWork
        Meshwork object with annotation properties:
        * `compartment`: SWC compartment labels for vertices (1: soma, 2: axon, 3: dendrite)
        * `is_axon`: Vertices with compartment label 2.
        * `lvl2_ids`: Level 2 id for each vertex of the spatial graph.
        * `pre_syn` and `post_syn` (optional): Synapse information.
        * `segment_properties` and `volume_properties` (optional): Segment and volume information.

        Note that the spatial graph (`nrn.mesh`) part of the file will _not_ be correct with this
        function unless `restore_graph` is set to True, but the skeleton and annotations will be.
    """
    sk = client.skeleton.get_skeleton(
        root_id, skeleton_version=skeleton_version, output_format="dict"
    )
    ts = client.chunkedgraph.get_root_timestamps(root_id, latest=True)[0]
    if synapse_reference_tables is None:
        synapse_reference_tables = {}
    return rebuild_meshwork(
        root_id=root_id,
        sk_verts=np.array(sk["vertices"]),
        sk_edges=np.array(sk["edges"]),
        root=int(sk["root"]),
        mesh_to_skel_map=np.array(sk["mesh_to_skel_map"]),
        lvl2_ids=np.array(sk["lvl2_ids"]),
        metadata=sk["meta"],
        radius=np.array(sk["radius"]),
        compartments=np.array(sk["compartment"]),
        client=client,
        synapses=synapses,
        timestamp=ts,
        restore_graph=restore_graph,
        restore_properties=restore_properties,
        synapse_reference_tables=synapse_reference_tables,
    )



import pathlib
import pickle

import numpy as np
import pandas as pd
import pytest
from pytest_mock import mocker
from caveclient.tools.testing import CAVEclientMock
from meshparty import meshwork
from packaging.version import Version

base_path = pathlib.Path(__file__).parent.resolve()

TEST_DATASTACK = "minnie65_public"
TEST_DATASTACK_NOCACHE = "minnie65_public_nocache"
MAT_VERSION = 795

INFO_CACHE = {
    TEST_DATASTACK: {
        "proofreading_review_table": None,
        "description": "This is the publicly released version of the minnie65 volume and segmentation. ",
        "proofreading_status_table": None,
        "cell_identification_table": None,
        "local_server": "https://minnie.microns-daf.com",
        "aligned_volume": {
            "description": "This is the second alignment of the IARPA 'minnie65' dataset, completed in the spring of 2020 that used the seamless approach.",
            "name": "minnie65_phase3",
            "id": 1,
            "image_source": "precomputed://https://bossdb-open-data.s3.amazonaws.com/iarpa_microns/minnie/minnie65/em",
            "display_name": "Minnie65",
        },
        "synapse_table": "synapses_pni_2",
        "viewer_site": "https://neuroglancer.neuvue.io",
        "analysis_database": None,
        "segmentation_source": "graphene://https://minnie.microns-daf.com/segmentation/table/minnie65_public",
        "soma_table": "nucleus_detection_v0",
        "viewer_resolution_x": 4.0,
        "viewer_resolution_y": 4.0,
        "viewer_resolution_z": 40.0,
    },
    TEST_DATASTACK_NOCACHE: {
        "proofreading_review_table": None,
        "description": "This is the publicly released version of the minnie65 volume and segmentation. ",
        "proofreading_status_table": None,
        "cell_identification_table": None,
        "local_server": "https://minnie.microns-daf.com",
        "aligned_volume": {
            "description": "This is the second alignment of the IARPA 'minnie65' dataset, completed in the spring of 2020 that used the seamless approach.",
            "name": "minnie65_phase3",
            "id": 1,
            "image_source": "precomputed://https://bossdb-open-data.s3.amazonaws.com/iarpa_microns/minnie/minnie65/em",
            "display_name": "Minnie65",
        },
        "synapse_table": "synapses_pni_2",
        "viewer_site": "https://neuroglancer.neuvue.io",
        "analysis_database": None,
        "segmentation_source": "graphene://https://minnie.microns-daf.com/segmentation/table/minnie65_public",
        "soma_table": "nucleus_detection_v0",
        "viewer_resolution_x": 4.0,
        "viewer_resolution_y": 4.0,
        "viewer_resolution_z": 40.0,
    },
}


@pytest.fixture
def root_id():
    return 864691136040432126


@pytest.fixture
def center_pt():
    return (177408, 157968, 21002)


class MockCloudVolume:
    @property
    def mip_resolution(self):
        return [4, 4, 40]


test_cv = MockCloudVolume()
with open(base_path / "data/l2_info.pkl", "rb") as f:
    test_l2ids = pickle.load(f)

test_l2graph = np.load(base_path / "data/lvl2_eg.npy")


test_pre_synapses = pd.read_feather(base_path / "data/pre_syn.feather")
test_post_synapses = pd.read_feather(base_path / "data/post_syn.feather")


@pytest.fixture()
def test_l2eg():
    return test_l2graph


@pytest.fixture()
def test_client(mocker):
    client = CAVEclientMock(
        datastack_name=TEST_DATASTACK,
        local_server="https://minnie.microns-daf.com",
        info_file=INFO_CACHE[TEST_DATASTACK],
        chunkedgraph=True,
        materialization=True,
        l2cache=True,
    )

    mocker.patch.object(client.info, "segmentation_cloudvolume", return_value=test_cv)
    mocker.patch.object(client.l2cache, "get_l2data", return_value=test_l2ids)
    mocker.patch.object(client.l2cache, "has_cache", return_value=True)
    mocker.patch.object(
        client.chunkedgraph, "level2_chunk_graph", return_value=test_l2graph
    )
    synapse_table = mocker.PropertyMock(return_value="synapses_pni_2")
    type(client.materialize).synapse_table = synapse_table
    mocker.patch.object(client.materialize, "get_versions", return_value=[1])
    mocker.patch.object(client.materialize, "get_timestamp", return_value="1234")
    mocker.patch(
        "pcg_skel.pcg_anno.get_level2_synapses",
        return_value=(test_pre_synapses, test_post_synapses),
    )
    return client


@pytest.fixture()
def test_client_nol2cache(mocker):
    client = CAVEclientMock(
        datastack_name=TEST_DATASTACK_NOCACHE,
        local_server="https://minnie.microns-daf.com",
        info_file=INFO_CACHE[TEST_DATASTACK_NOCACHE],
        l2cache=True,
        l2cache_disabled=True,
    )
    mocker.patch.object(
        client.materialize, "_get_version", return_value=Version("4.35.0")
    )
    mocker.patch.object(
        client.chunkedgraph, "_get_version", return_value=Version("2.17.2")
    )

    mocker.patch.object(client.info, "segmentation_cloudvolume", return_value=test_cv)
    mocker.patch.object(client.l2cache, "has_cache", return_value=False)
    mocker.patch.object(
        client.chunkedgraph, "level2_chunk_graph", return_value=test_l2graph
    )
    return client


@pytest.fixture()
def test_neuron():
    return meshwork.load_meshwork(base_path / "data" / "test_neuron.h5")



import cloudvolume
import fastremap
import multiwrapper.multiprocessing_utils as mu
import numpy as np
from scipy import spatial

from . import chunk_cache, utils

UnshardedMeshSource = (
    cloudvolume.datasource.graphene.mesh.unsharded.GrapheneUnshardedMeshSource
)
ShardedMeshSource = (
    cloudvolume.datasource.graphene.mesh.sharded.GrapheneShardedMeshSource
)

L2_SERVICE_NAME = "service"


class CompleteDataException(Exception):
    pass

def build_graph_topology(lvl2_edge_graph):
    """Extract the graph from the result of the lvl2_graph endpoint"""
    lvl2_edge_graph = np.unique(np.sort(lvl2_edge_graph, axis=1), axis=0)
    lvl2_ids = np.unique(lvl2_edge_graph)
    l2dict = {l2id: ii for ii, l2id in enumerate(lvl2_ids)}
    eg_arr_rm = fastremap.remap(lvl2_edge_graph, l2dict)
    l2dict_reversed = {ii: l2id for l2id, ii in l2dict.items()}
    return eg_arr_rm, l2dict, l2dict_reversed, lvl2_ids


def build_spatial_graph(
    lvl2_edge_graph, cv=None, client=None, method="chunk", require_complete=False
):
    """Extract spatial graph and level 2 id lookups from chunkedgraph "lvl2_graph" endpoint.

    Parameters
    ----------
    lvl2_edge_graph : array
        Nx2 edge list of level 2 ids
    cv : cloudvolume.CloudVolume
        Associated cloudvolume object

    Returns
    -------
    eg_arr_rm : np.array
        Nx2 edge list of indices remapped to integers starting at 0 through M, the number of unique level 2 ids.
    l2dict : dict
        Dict with level 2 ids as keys and vertex index as values
    l2dict_reversed : dict
        Dict with vertex index as keys and level 2 id as values
    x_ch : np.array
        Mx3 array of vertex locations in chunk index space.
    """
    eg_arr_rm, l2dict, l2dict_reversed, lvl2_ids = build_graph_topology(lvl2_edge_graph)
    if method == "chunk":
        if cv is None and client is None:
            raise ValueError("Either cv or client must be provided")
        if cv is None:
            cv = client.info.segmentation_cloudvolume(progress=False)
        x_ch = [np.array(cv.mesh.meta.meta.decode_chunk_position(l)) for l in lvl2_ids]
    elif method == "service":
        x_ch = dense_spatial_lookup(
            lvl2_ids,
            eg_arr_rm,
            client,
            require_complete=require_complete,
        )
    return eg_arr_rm, l2dict, l2dict_reversed, x_ch

def adjust_meshwork(nrn, cv):
    """Transform vertices in chunk index space to euclidean"""
    nrn._mesh.vertices = utils.chunk_to_nm(nrn._mesh.vertices, cv)


def dense_spatial_lookup(l2ids, eg, client, require_complete=False):
    l2means = np.full((len(l2ids), 3), np.nan)
    locs, inds_found = chunk_cache.get_locs_remote(l2ids, client)
    if require_complete:
        if not np.all(inds_found):
            raise CompleteDataException("Some chunk indices are not yet computed")
    l2means[inds_found] = locs
    return l2means


def refine_vertices(
    vertices,
    l2dict_reversed,
    cv,
    refine_inds="all",
    scale_chunk_index=True,
    convert_missing=False,
    return_missing_ids=True,
    cache=None,
    save_to_cache=False,
    segmentation_fallback=True,
    fallback_mip=2,
    client=None,
):
    """Refine vertices in chunk index space by converting to euclidean space using a combination of mesh downloading and simple chunk mapping.

    Parameters
    ----------
    vertices : array
        Nx3 array of vertex locations in chunk index space
    l2dict_reversed : dict or array
        N-length mapping from vertex index to uint64 level 2 id.
    cv : cloudvolume.CloudVolume
        CloudVolume associated with the chunkedgraph instance
    refine_inds : array, string, or None, optional
        Array of indices to refine via mesh download and recentering, None, or 'all'. If 'all', does all vertices. By default 'all'.
    scale_chunk_index : bool, optional
        If True, moves chunk indices to the euclidean space (the center of the chunk) if not refined. by default True.
    convert_missing : bool, optional
        If True, vertices with missing meshes are converted to the center of their chunk. Otherwise, they are given nans. By default, False.
    return_missing_ids : bool, optional
        If True, returns a list of level 2 ids for which meshes were not found, by default True
    segmentation_fallback : bool, optional
        If True, uses the segmentation as a fallback if the mesh does not exist.

    Returns
    -------
    new_vertices : array
        Nx3 array of remapped vertex locations in euclidean space
    missing_ids : array, optional
        List of level 2 ids without meshes. Only returned if return_missing_ids is True.
    """
    vertices = vertices.copy()

    if isinstance(refine_inds, str):
        if refine_inds == "all":
            refine_inds = np.arange(0, len(vertices))
        else:
            raise ValueError("Invalid value for refine_inds")

    if refine_inds is not None:
        l2ids = [l2dict_reversed[k] for k in refine_inds]
        pt_locs, missing_ids = lvl2_fragment_locs(
            l2ids,
            cv,
            return_missing=True,
            cache=cache,
            save_to_cache=save_to_cache,
            segmentation_fallback=segmentation_fallback,
            fallback_mip=fallback_mip,
            client=client,
        )

        if convert_missing:
            missing_inds = np.any(np.isnan(pt_locs), axis=1)
            vertices[refine_inds[~missing_inds]] = pt_locs[~missing_inds]
        else:
            missing_inds = np.full(len(pt_locs), False)
            vertices[refine_inds] = pt_locs
    else:
        refine_inds = np.array([], dtype=int)
        missing_inds = np.array([], dtype=bool)
        missing_ids = np.array([], dtype=int)

    if scale_chunk_index and len(refine_inds) != len(vertices):
        # Move unrefined vertices to center of chunks
        other_inds = np.full(len(vertices), True)
        if len(refine_inds) > 0:
            other_inds[refine_inds[~missing_inds]] = False
        vertices[other_inds] = (
            utils.chunk_to_nm(vertices[other_inds], cv) + utils.chunk_dims(cv) // 2
        )
    if return_missing_ids:
        return vertices, missing_ids
    else:
        return vertices


def get_root_id_from_point(point, voxel_resolution, client):
    cv = cloudvolume.CloudVolume(
        client.info.segmentation_source(),
        use_https=True,
        bounded=False,
        fill_missing=True,
        progress=False,
        secrets={"token": client.auth.token},
    )
    return int(
        cv.download_point(
            point, size=1, coord_resolution=voxel_resolution, agglomerate=True
        )
    )


def get_closest_lvl2_chunk(
    point,
    root_id,
    client,
    cv=None,
    voxel_resolution=None,
    radius=200,
    return_point=False,
):
    """Get the closest level 2 chunk on a root id

    Parameters
    ----------
    point : array-like
        Point in voxel space.
    root_id : int
        Root id of the object
    client : CAVEclient
        CAVE client to access data
    cv : cloudvolume.CloudVolume or None, optional
        Cloudvolume associated with the dataset. One is created if None.
    voxel_resolution : list, optional
        Point resolution to map between point resolution and mesh resolution, by default [4, 4, 40]
    radius : int, optional
        Max distance to look for a nearby supervoxel. Optional, default is 200.
    return_point : bool, optional
        If True, returns the closest point in addition to the level 2 id. Optional, default is False.

    Returns
    -------
    level2_id : int
        Level 2 id of the object nearest to the point specified.
    close_point : array, optional
        Closest point inside the object to the specified point. Only returned if return_point is True.
    """
    if cv is None:
        cv = client.info.segmentation_cloudvolume(progress=False)

    if voxel_resolution is None:
        voxel_resolution = np.array(cv.mip_resolution(0))

    point = point * np.array(voxel_resolution)
    # Get the closest adjacent point for the root id within the radius.
    mip_scaling = np.array(cv.mip_resolution(0))

    pt = np.array(np.array(point) // mip_scaling, np.int32)
    offset = np.array(radius // mip_scaling, np.int32)
    lx = np.array(pt) - offset
    ux = np.array(pt) + offset
    bbox = cloudvolume.Bbox(lx.astype(int), ux.astype(int))
    
    ts = client.chunkedgraph.get_root_timestamps(root_id, latest=True)[0]

    vol = cv.download(bbox, segids=[root_id], timestamp=ts)
    vol = np.squeeze(vol)
    if not bool(np.any(vol > 0)):
        raise ValueError("No point of the root id is near the specified point")

    ctr = offset * mip_scaling  # Offset is at center of the volume.
    xyz = np.vstack(np.where(vol > 0)).T
    xyz_nm = xyz * mip_scaling

    ind = np.argmin(np.linalg.norm(xyz_nm - ctr, axis=1))
    closest_pt = vol.bounds.minpt + xyz[ind]

    # Look up the level 2 supervoxel for that id.
    closest_sv = int(cv.download_point(closest_pt, size=1))
    lvl2_id = int(client.chunkedgraph.get_root_id(closest_sv, level2=True))

    if return_point:
        return lvl2_id, closest_pt * mip_scaling
    else:
        return lvl2_id


def lvl2_fragment_locs(
    l2_ids,
    cv=None,
    return_missing=True,
    segmentation_fallback=True,
    fallback_mip=2,
    cache=None,
    save_to_cache=False,
    client=None,
):
    """Look up representitive location for a list of level 2 ids.

    The representitive point for a mesh is the mesh vertex nearest to the
    centroid of the mesh fragment.

    Parameters
    ----------
    l2_ids : list-like
        List of N level 2 ids
    cv : cloudvolume.CloudVolume
        Associated cloudvolume object
    return_missing : bool, optional
        If True, returns ids of missing meshes. Default is True
    segmentation_fallback : bool, optional
        If True, uses segmentation to get the location if no mesh fragment is found.
        This is slower, but more robust. Default is True.
cache: str or None, optional
        If a string, filename for a sqlite database used as a lookup cache for l2 ids.
        Default is None.
    save_to_cache: bool, optional
        If True and a chace is set, automatically saves looked up locations to the cache.
        Default is False.
    client : CAVEclient, optional
        Client to be used if the remote service is being used.

    Returns
    -------
    l2means : np.array
        Nx3 list of point locations. Missing mesh fragments get a nan for each component.
    missing_ids : np.array
        List of level 2 ids that were not found. Only returned if return_missing is True.
    """

    l2_ids = np.array(l2_ids)
    l2means = np.full((len(l2_ids), 3), np.nan, dtype=float)
    if cache is not None:
        if cache == L2_SERVICE_NAME:
            l2means_cached, is_cached = chunk_cache.lookup_cached_ids(
                l2_ids, remote_cache=True, client=client
            )
        else:
            l2means_cached, is_cached = chunk_cache.lookup_cached_ids(
                l2_ids, remote_cache=False, cache_file=cache
            )
    else:
        l2means_cached = np.zeros((0, 3), dtype=float)
        is_cached = np.full(len(l2_ids), False, dtype=np.bool)

    l2means[is_cached] = l2means_cached

    if np.any(~is_cached):
        l2means_dl, missing_ids = download_lvl2_locs(
            l2_ids[~is_cached], cv, segmentation_fallback, fallback_mip
        )
        l2means[~is_cached] = l2means_dl
        if cache is not None and save_to_cache and cache != L2_SERVICE_NAME:
            chunk_cache.save_ids_to_cache(l2_ids[~is_cached], l2means_dl, cache)
    else:
        missing_ids = []

    if return_missing:
        return l2means, missing_ids
    else:
        return l2means


def download_lvl2_locs(l2_ids, cv, segmentation_fallback, fallback_mip=2):
    l2meshes = download_l2meshes(
        l2_ids, cv, sharded=isinstance(cv.mesh, ShardedMeshSource)
    )

    l2means = []
    args = []
    auth_token = cv.meta.auth_header["Authorization"][len("Bearer ") :]
    for l2id in l2_ids:
        args.append(
            (
                l2id,
                l2meshes.get(l2id, None),
                f"graphene://{cv.meta.table_path}",
                auth_token,
                segmentation_fallback,
                fallback_mip,
            )
        )

    l2means = mu.multiprocess_func(_localize_l2_id_multi, args, n_threads=1)

    if len(l2means) > 0:
        l2means = np.vstack(l2means)
        missing_ids = l2_ids[np.isnan(l2means[:, 0])]
    else:
        l2means = np.empty((0, 3), dtype=float)
        missing_ids = []
    return l2means, missing_ids


def _localize_l2_id_multi(args):
    l2id, l2mesh, cv_path, auth_token, segmentation_fallback, fallback_mip = args
    return _localize_l2_id(
        l2id, l2mesh, cv_path, auth_token, segmentation_fallback, fallback_mip
    )


def _localize_l2_id(
    l2id, l2mesh, cv_path, auth_token, segmentation_fallback, fallback_mip
):
    if l2mesh is not None:
        l2m_abs = np.mean(l2mesh.vertices, axis=0)
        _, ii = spatial.cKDTree(l2mesh.vertices).query(l2m_abs)
        l2m = l2mesh.vertices[ii]
    else:
        if segmentation_fallback:
            cv = cloudvolume.CloudVolume(
                cv_path,
                bounded=False,
                progress=False,
                fill_missing=True,
                use_https=True,
                mip=0,
                secrets={"token": auth_token},
            )
            try:
                l2m = chunk_location_from_segmentation(l2id, cv, mip=fallback_mip)
            except:
                l2m = np.array([np.nan, np.nan, np.nan])
            del cv
        else:
            l2m = np.array([np.nan, np.nan, np.nan])
    return l2m


def download_l2meshes(l2ids, cv, sharded=False):
    cv.parallel = 1
    if sharded:
        return cv.mesh.get_meshes_on_bypass(l2ids, allow_missing=True)
    else:
        return cv.mesh.get(
            l2ids, allow_missing=True, deduplicate_chunk_boundaries=False
        )


def chunk_location_from_segmentation(l2id, cv, mip=0):
    """Representative level 2 id location using the segmentation.
    This is typically slower than the mesh approach, but is more robust.

    Parameters
    ----------
    l2id : int
        Level 2 id to look up
    cv : cloudvolume.CloudVolume
        CloudVolume associated with the mesh

    Returns
    -------
    xyz_rep : np.array
        3-element xyz array of the representative location in euclidean space.
    """
    loc_ch = np.array(cv.mesh.meta.meta.decode_chunk_position(l2id))
    loc_vox = np.atleast_2d(loc_ch) * cv.graph_chunk_size + np.array(cv.voxel_offset)

    bbox = cloudvolume.Bbox(loc_vox[0], loc_vox[0] + cv.graph_chunk_size)
    bbox_mip = cv.bbox_to_mip(bbox, 0, mip)

    try:
        sv_vol, _ = cv.download(bbox_mip, segids=(l2id,), mip=mip, renumber=True)
        sv_vol = sv_vol > 0
        x, y, z = np.where(sv_vol.squeeze())

        if len(x) == 0 and mip > 0:
            # If too small for the requested mip level, jump to the highest mip level
            del sv_vol
            return chunk_location_from_segmentation(l2id, cv, mip=0)
    except:
        return np.array([np.nan, np.nan, np.nan])

    xyz = np.vstack((x, y, z)).T * np.array(sv_vol.resolution)

    xyz_mean = np.mean(xyz, axis=0)
    xyz_box = xyz[np.argmin(np.linalg.norm(xyz - xyz_mean, axis=1))]
    xyz_rep = np.array(sv_vol.bounds.minpt) * np.array(sv_vol.resolution) + xyz_box
    del sv_vol
    return xyz_rep



from __future__ import annotations
import datetime
from caveclient.frameworkclient import CAVEclientFull
from typing import Union, Optional
import pandas as pd
from typing import Optional, Dict, Any, Tuple, Union, List


def annotation_to_level2_id(
    df: pd.DataFrame,
    client: CAVEclientFull,
    bound_pt_columns: str = "pt",
    l2_suffix: str = "_level2_id",
    sv_columns: Optional[Union[str, list[str]]] = None,
    l2_columns: Optional[Union[str, list[str]]] = None,
    inplace: bool = False,
    timestamp: Optional[datetime.datetime] = None,
) -> pd.DataFrame:
    """Add or more level2_id columns to a dataframe based on supervoxel columns

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with one or more supervoxel columns
    client : CAVEclient or ChunkedgraphClient
        Client for interacting with the chunkedgraph
    bound_pt_columns : str or list-like, optional
        List of bound spatial point names, prefix only. E.g. 'pt' for an annotation with
        'pt_position', 'pt_supervoxel_id', and 'pt_root_id'. Optional, by default 'pt'.
    l2_suffix : str, optional
        Suffix to use for new level 2 id column, by default '_level2_id'
    sv_columns : str or list-like, optional
        Explicit list of level 2 columns, not needed if using bound_pt_columns. By default None
    l2_columns : [type], optional
        Explicit list of desired level 2 columns, not needed if using bound_pt_columns. By default None
    inplace : bool, optional
        If True, change the dataframe in place, by default False
    timestamp : datetime or None, optional
        Timestamp to lookup the mapping. If None, uses the materialization version timestamp.
        Default is None.

    Returns
    -------
    pd.DataFrame
        DataFrame with level 2 id columns added
    """
    if isinstance(bound_pt_columns, str):
        bound_pt_columns = [bound_pt_columns]
    if sv_columns is None:
        sv_columns = [f"{c}_supervoxel_id" for c in bound_pt_columns]
        if l2_columns is None:
            l2_columns = [f"{c}{l2_suffix}" for c in bound_pt_columns]
    elif l2_columns is None:
        l2_columns = [f"{c}{l2_suffix}" for c in sv_columns]

    if timestamp is None:
        timestamp = client.materialize.get_timestamp()

    if isinstance(client, CAVEclientFull):
        pcg_client = client.chunkedgraph
    else:
        pcg_client = client

    if not inplace:
        df = df.copy()

    for col, l2_col in zip(sv_columns, l2_columns):
        level2_ids = pcg_client.get_roots(
            df[col].values,
            stop_layer=2,
            timestamp=timestamp,
        )
        df[l2_col] = level2_ids
    return df


def annotation_to_mesh_index(
    df: pd.DataFrame,
    l2dict: dict,
    level2_id_col: Union[str, list[str]] = "pt_level2_id",
    mesh_index_col: str = "pt_mesh_ind",
    inplace: bool = False,
) -> pd.DataFrame:
    """Map level2 ids to mesh indices.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with at least one level 2 id column
    l2dict : dict
        Dict of level2 id to mesh index mappings
    level2_id_col : str or list, optional
        Level 2 id column or list of columns, by default 'pt_level2_id'
    mesh_index_col : str, optional
        Column name (or names) to use for added mesh index, by default 'pt_mesh_ind'
    inplace : bool, optional
        If False, makes changes on a copy of the dataframe, by default False

    Returns
    -------
    pandas.DataFrame
        DataFrame with mesh index column/s added
    """
    if not inplace:
        df = df.copy()
    if isinstance(level2_id_col, str):
        level2_id_col = [level2_id_col]
    if isinstance(mesh_index_col, str):
        mesh_index_col = [mesh_index_col]

    for l2col, mind_col in zip(level2_id_col, mesh_index_col):
        df[mind_col] = df[l2col].apply(lambda x: l2dict[x])
    return df


def get_level2_synapses(
    root_id: int,
    l2dict: Dict[Any, Any],
    client: CAVEclientFull,
    synapse_table: str,
    remove_self: bool = True,
    pre: bool = True,
    post: bool = True,
    live_query: bool = False,
    timestamp: Optional[datetime.datetime] = None,
    metadata: bool = False,
    synapse_point_resolution: Optional[List[float]] = None,
    synapse_reference_tables: dict = {},
) -> Tuple[Optional[pd.DataFrame], Optional[pd.DataFrame]]:
    """Retrieve level 2 synapses for a given root ID.

    Parameters
    ----------
    root_id : int
        The root ID for which to retrieve synapses.
    l2dict : dict
        Dict of level2 id to mesh index mappings
    client : object
        The client object used for querying the synapse data.
    synapse_table : str
        The name of the synapse table to query.
    remove_self : bool, optional
        Whether to remove self-synapses. Defaults to True.
    pre : bool, optional
        Whether to retrieve pre-synapses. Defaults to True.
    post : bool, optional
        Whether to retrieve post-synapses. Defaults to True.
    live_query : bool, optional
        Whether to perform a live query. Defaults to False.
    timestamp : datetime, optional
        The timestamp for the live query. Defaults to None.
    metadata : bool, optional
        Whether to include metadata in the query results. Defaults to False.
    synapse_point_resolution : list, optional
        The resolution of the synapse points. Defaults to None.
    synapse_reference_tables : dict, optional
        A dictionary of synapse reference tables to attach to synapses.
        Keys are 'pre' and 'post', values are table names. Defaults to {}.

    Returns
    -------
    pre_syn_df : DataFrame or None
        The DataFrame containing pre-synapse data, or None if pre is False.
    post_syn_df : DataFrame or None
        The DataFrame containing post-synapse data, or None if post is False.
    """
    live_query = timestamp is not None

    if timestamp is None:
        timestamp = datetime.datetime.now(datetime.timezone.utc)
    if pre is True:
        pre_syn_df = _mapped_synapses(
            root_id,
            client,
            l2dict,
            "pre",
            synapse_table,
            remove_self=remove_self,
            live_query=live_query,
            timestamp=timestamp,
            metadata=metadata,
            synapse_point_resolution=synapse_point_resolution,
            synapse_reference_tables=synapse_reference_tables,
        )
    else:
        pre_syn_df = None

    if post is True:
        post_syn_df = _mapped_synapses(
            root_id,
            client,
            l2dict,
            "post",
            synapse_table,
            remove_self=remove_self,
            live_query=live_query,
            timestamp=timestamp,
            metadata=metadata,
            synapse_point_resolution=synapse_point_resolution,
            synapse_reference_tables=synapse_reference_tables,
        )
    else:
        post_syn_df = None

    return pre_syn_df, post_syn_df


def _mapped_synapses(
    root_id,
    client,
    l2dict,
    side,
    synapse_table,
    remove_self,
    live_query,
    timestamp,
    metadata,
    remove_crud=True,
    synapse_point_resolution=None,
    synapse_reference_tables={},
):
    if timestamp:
        syn_df = client.materialize.live_live_query(
            synapse_table,
            filter_equal_dict={synapse_table: {f"{side}_pt_root_id": root_id}},
            timestamp=timestamp,
            metadata=metadata,
            desired_resolution=synapse_point_resolution,
            log_warning=False,
        )
        if isinstance(synapse_reference_tables.get(side), str):
            synapse_reference_tables[side] = [synapse_reference_tables[side]]
        for tbl in synapse_reference_tables.get(side, []):
            if len(syn_df)>0:
                aug_df = client.materialize.live_live_query(
                    tbl,
                    filter_in_dict={tbl: {'target_id': syn_df['id'].values.tolist()}},
                    timestamp=timestamp,
                    log_warning=False,
                )
                aug_df.drop(columns=["created", "superceded_id", "valid", "id"], inplace=True, errors="ignore")
                syn_df = syn_df.merge(aug_df.rename(columns={'target_id': 'id'}), on='id', how='left')
    else:
        syn_df = client.materialize.query_table(
            synapse_table,
            filter_equal_dict={f"{side}_pt_root_id": root_id},
            metadata=metadata,
            desired_resolution=synapse_point_resolution,
        )

    if remove_crud:
        syn_df.drop(
            columns=["created", "superceded_id", "valid"], inplace=True, errors="ignore"
        )

    if remove_self:
        syn_df = syn_df.query("pre_pt_root_id != post_pt_root_id").reset_index(
            drop=True
        )
    syn_df = annotation_to_level2_id(
        syn_df,
        client,
        bound_pt_columns=f"{side}_pt",
        inplace=True,
        timestamp=timestamp,
    )
    syn_df = annotation_to_mesh_index(
        syn_df,
        l2dict,
        level2_id_col=f"{side}_pt_level2_id",
        mesh_index_col=f"{side}_pt_mesh_ind",
        inplace=True,
    )
    return syn_df



import warnings

import numpy as np


def filter_l2dict(sk_ch, l2dict_mesh_r):
    """Converts a reverse level 2 dict for a mesh into one for a skeleton"""
    l2dict_sk_r = {
        ii: l2dict_mesh_r.get(mind, -1) for ii, mind in enumerate(sk_ch.mesh_index)
    }
    l2dict_sk = {k: v for v, k in l2dict_sk_r.items()}
    return l2dict_sk, l2dict_sk_r


def propagate_l2dict(sk_ch, l2dict_mesh):
    """Propagate a l2dict for a mesh to an l2dict for a skeleton, with multiple
    level 2 ids potentially mapping to each skeleton index
    """
    l2dict_long = {k: sk_ch.mesh_to_skel_map[v] for k, v in l2dict_mesh.items()}
    return l2dict_long


def fix_nan_verts_mesh(mesh, num_rounds=20):
    if num_rounds is None:
        num_rounds = 0
    # graph = mesh.csgraph.todense() > 1

    nr = 0
    while nr < num_rounds:
        nanvinds = np.flatnonzero(np.isnan(mesh.vertices[:, 0]))
        if len(nanvinds) == 0:
            return

        new_verts = mesh.vertices[nanvinds]

        for ii, vid in enumerate(nanvinds):
            neib_inds = mesh.csgraph[vid, :].indices
            if len(neib_inds) == 0:
                continue
            if np.any(neib_inds):
                is_single = int(len(neib_inds) == 1)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    new_verts[ii] = (
                        np.nanmean(mesh.vertices[neib_inds], axis=0)
                        + np.array([0, 0, 1]) * is_single
                    )
        mesh.vertices[nanvinds] = new_verts

        nr += 1
    else:
        if np.any(np.isnan(mesh.vertices)):
            print(f"Could not fix all nans after {num_rounds} rounds")


def fix_nan_verts(sk, num_rounds=20):
    """Replace vertices with locations that are nan with mean of neighbor locations"""
    if num_rounds is None:
        num_rounds = 0

    nr = 0
    while nr < num_rounds:
        nanvinds = np.flatnonzero(np.isnan(sk.vertices[:, 0]))
        new_verts = sk.vertices[nanvinds]

        for ii, vid in enumerate(nanvinds):
            neib_inds = (
                np.array(sk.csgraph_binary_undirected[vid, :].todense()).squeeze() > 0
            )
            if len(neib_inds) == 0:
                continue
            if np.any(neib_inds):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    new_verts[ii] = np.nanmean(
                        sk.vertices[neib_inds], axis=0
                    ) + np.array([0, 0, 1])
        sk._rooted._vertices[nanvinds] = new_verts
        sk._vertices[nanvinds] = new_verts

        if not np.any(np.isnan(new_verts[:, 0])):
            break

        nr += 1
    else:
        if np.any(np.isnan(sk.vertices)):
            print(f"Could not fix all nans after {num_rounds} rounds")


def get_centered_mesh(mesh):
    if mesh is None:
        return np.array([np.nan, np.nan, np.nan])
    verts = mesh.vertices
    close_ind = np.argmin(np.linalg.norm(verts - np.mean(verts, axis=0), axis=1))
    return verts[close_ind]


def attach_new_skeleton(nrn, new_skeleton):
    nrn._skeleton = new_skeleton
    nrn._recompute_indices()



import pytest
import pcg_skel
from pcg_skel.chunk_cache import NoL2CacheException
import numpy as np

from conftest import test_client, test_client_nol2cache
from conftest import root_id, center_pt


def test_pcg_graph(test_client, root_id):
    graph_m = pcg_skel.pcg_graph(root_id, test_client)
    assert graph_m.vertices is not None
    assert graph_m.edges is not None


def test_pcg_skeleton(test_client, root_id, center_pt):
    graph_sk = pcg_skel.pcg_skeleton(
        root_id,
        test_client,
        collapse_radius=True,
        root_point=center_pt,
        root_point_resolution=[4, 4, 40],
    )
    assert graph_sk.vertices is not None
    assert graph_sk.edges is not None
    assert graph_sk.path_length() > 0


def test_pcg_skeleton_prebaked(test_client, root_id, center_pt, test_l2eg):
    graph_sk = pcg_skel.pcg_skeleton(
        root_id,
        test_client,
        collapse_radius=True,
        root_point=center_pt,
        root_point_resolution=[4, 4, 40],
        level2_graph=test_l2eg,
    )
    assert graph_sk.vertices is not None
    assert graph_sk.edges is not None
    assert graph_sk.path_length() > 0


def test_pcg_skeleton_direct(test_client, root_id, center_pt):
    graph_m = pcg_skel.pcg_graph(root_id, test_client)
    sk = pcg_skel.pcg_skeleton_direct(
        vertices=graph_m.vertices,
        edges=graph_m.graph_edges,
    )
    assert len(sk.vertices) == 2567


def test_pcg_meshwork(test_client, root_id, center_pt):
    nrn = pcg_skel.pcg_meshwork(
        root_id=root_id,
        client=test_client,
        collapse_soma=True,
        root_point=center_pt,
        root_point_resolution=[4, 4, 40],
        synapses="all",
        synapse_table=test_client.materialize.synapse_table,
    )
    assert len(nrn.anno.table_names) > 0
    assert nrn.mesh.vertices.shape[0] > 0


def test_defunct_meshwork(test_client, root_id, center_pt):
    nrn2 = pcg_skel.coord_space_meshwork(
        root_id=root_id,
        client=test_client,
        collapse_soma=True,
        root_point=center_pt,
        root_point_resolution=[4, 4, 40],
        synapses="all",
        synapse_table=test_client.materialize.synapse_table,
    )
    assert len(nrn2.anno.table_names) > 0
    assert nrn2.mesh.vertices.shape[0] > 0


def test_pcg_meshwork_noclient(test_client_nol2cache, root_id, center_pt):
    with pytest.raises(NoL2CacheException):
        pcg_skel.pcg_skeleton(
            root_id,
            test_client_nol2cache,
            collapse_radius=True,
            root_point=center_pt,
            root_point_resolution=[4, 4, 40],
        )


def test_feature_aggregation(test_neuron):
    orig_df = test_neuron.anno["synapse_count"].df
    skel_df = pcg_skel.features.aggregate_property_to_skeleton(
        test_neuron,
        "synapse_count",
        agg_dict={
            "num_syn_out": "sum",
            "num_syn_in": "sum",
            "net_size_out": "sum",
            "net_size_in": "sum",
        },
    )
    for col in ["num_syn_in", "num_syn_out", "net_size_in", "net_size_out"]:
        assert skel_df.sum()[col] == orig_df.sum()[col]



from __future__ import annotations
from . import chunk_tools, features
from . import skel_utils as sk_utils

import cloudvolume
import warnings
import numpy as np
import datetime

from caveclient import CAVEclient
from meshparty import meshwork, skeletonize, trimesh_io
from typing import Union, Optional

Numeric = Union[int, float, np.number]

DEFAULT_VOXEL_RESOLUTION = [4, 4, 40]
DEFAULT_COLLAPSE_RADIUS = 7500.0
DEFAULT_INVALIDATION_D = 7500

skeleton_type = "pcg_skel"


def pcg_graph(
    root_id: int,
    client: CAVEclient.frameworkclient.CAVEclientFull,
    cv: cloudvolume.CloudVolume = None,
    return_l2dict: bool = False,
    nan_rounds: int = 10,
    require_complete: bool = False,
    level2_graph: Optional[np.ndarray] = None,
):
    """Compute the level 2 spatial graph (or mesh) of a given root id using the l2cache.

    Parameters
    ----------
    root_id : int
        Root id of a segment
    client : CAVEclient.caveclient
        Initialized CAVEclient for the dataset.
    cv : cloudvolume.CloudVolume
        Initialized CloudVolume object for the dataset. This does not replace the caveclient, but
        a pre-initizialized cloudvolume can save some time during batch processing.
    return_l2dict : bool
        If True, returns the mappings between l2 ids and vertices.
    nan_rounds : int
        If vertices are missing (or not computed), this sets the number of iterations for smoothing over them.
    require_complete : bool
        If True, raise an Exception if any vertices are missing from the cache.
    level2_graph : np.ndarray, optional
        Level 2 graph for the root id as returned by client.chunkedgraph.level2_chunk_graph.
        A list of lists of edges between level 2 chunks, as defined by their chunk ids.
        If None, will query the chunkedgraph for the level 2 graph. Optional, by default None.


    Returns
    -------
    mesh : meshparty.trimesh_io.Mesh
        Object with a vertex for every level 2 id and edges between all connected level 2 chunks.
    l2dict : dict, optional
        Dictionary with keys as level 2 ids and values as mesh vertex index. Optional, only returned if `return_l2dict` is True.
    l2dict_reverse : dict, optional
        Dictionary with keys as mesh vertex indices and values as level 2 id. Optional, only returned if `return_l2dict` is True.
    """
 
    if level2_graph is None:
        lvl2_eg = client.chunkedgraph.level2_chunk_graph(root_id)
    else:
        lvl2_eg = level2_graph
    if len(lvl2_eg) == 0:
        l2id = client.chunkedgraph.get_leaves(root_id, stop_layer=2)
        lvl2_eg = np.array([l2id, l2id]).reshape((1, 2))
        is_singlet = True
    else:
        is_singlet = False

    eg, l2dict_mesh, l2dict_r_mesh, x_ch = chunk_tools.build_spatial_graph(
        lvl2_eg,
        cv=cv,
        client=client,
        method="service",
        require_complete=require_complete,
    )
    if is_singlet:
        eg = None
    mesh_loc = trimesh_io.Mesh(
        vertices=x_ch,
        faces=[[0, 0, 0]],  # Some functions fail if no faces are set.
        link_edges=eg,
    )

    sk_utils.fix_nan_verts_mesh(mesh_loc, nan_rounds)

    if return_l2dict:
        return mesh_loc, l2dict_mesh, l2dict_r_mesh
    else:
        return mesh_loc


def pcg_skeleton_direct(
    vertices,
    edges,
    invalidation_d: Numeric = DEFAULT_INVALIDATION_D,
    root_point: list = None,
    collapse_soma: bool = False,
    collapse_radius: Numeric = DEFAULT_COLLAPSE_RADIUS,
    root_id: Optional[int] = None,
):
    """
    Produce a skeleton from an already-computed l2graph.
    This is effectively a wrapper for meshparty skeletonize with a consistent set of parameters and format.

    Parameters
    ----------
    vertices : np.array
        Array of vertices for the mesh graph.
    edges : np.array
        Array of edges for the mesh graph.
    invalidation_d : int, optional
        Distance (in nanometers) for TEASAR skeleton invalidation.
    root_point : np.array, optional
        3-element list or array with the x,y,z location of the root point in same units as vertices.
        If None, the most distant tip is set to root.
    collapse_soma : bool, optional
        If True, collapse nearby vertices into the root point.
    collapse_radius : int, optional
        Distance (in nanometers) for soma collapse.
    root_id : int, optional
        Root id of the segment, used in metadata. Optional, by default None.
    level2_graph : np.ndarray, optional
        Level 2 graph for the root id as returned by client.chunkedgraph.level2_chunk_graph.
        A list of lists of edges between level 2 chunks, as defined by their chunk ids.
        If None, will query the chunkedgraph for the level 2 graph. Optional, by default None.


    Returns
    -------
    sk : meshparty.skeleton.Skeleton
        Skeleton for the l2graph.
    """

    l2graph = trimesh_io.Mesh(
        vertices=vertices,
        faces=[[0, 0, 0]],  # Some functions fail if no faces are set.
        link_edges=edges,
    )

    sk = skeletonize.skeletonize_mesh(
        l2graph,
        invalidation_d=invalidation_d,
        soma_pt=root_point,
        collapse_soma=collapse_soma,
        soma_radius=collapse_radius,
        compute_radius=False,
        cc_vertex_thresh=0,
        remove_zero_length_edges=True,
        meta={
            "root_id": root_id,
            "skeleton_type": skeleton_type,
            "meta": {"space": "l2cache", "datastack": None},
        },
    )
    return sk


def pcg_skeleton(
    root_id: int,
    client: CAVEclient.frameworkclient.CAVEclientFull,
    datastack_name: str = None,
    cv: cloudvolume.CloudVolume = None,
    invalidation_d: Numeric = 7500,
    return_mesh: bool = False,
    return_l2dict: bool = False,
    return_l2dict_mesh: bool = False,
    root_point: list = None,
    root_point_resolution: list = None,
    collapse_soma: bool = False,
    collapse_radius: Numeric = 7500,
    nan_rounds: int = 10,
    require_complete: bool = False,
    level2_graph: Optional[np.ndarray] = None,
):
    """Produce a skeleton from the level 2 graph.
    Parameters
    ----------
    root_id : int
        Root id of a segment
    client : CAVEclient.caveclient.CAVEclientFull
        Initialized CAVEclient for the dataset.
    datastack_name : string, optional
        If client is None, initializes a CAVEclient at this datastack.
    cv : cloudvolume.CloudVolume, optional
        Initialized CloudVolume object for the dataset. This does not replace the caveclient, but
        a pre-initizialized cloudvolume can save some time during batch processing.
    invalidation_d : int, optional
        Distance (in nanometers) for TEASAR skeleton invalidation.
    return_mesh : bool, optional
        If True, returns the mesh graph as well as the skeleton.
    return_l2dict : bool, optional
        If True, returns the mappings between l2 ids and skeleton vertices.
    return_l2dict_mesh : bool, optional
        If True, returns mappings between l2 ids and mesh graph vertices.
    root_point : npt.ArrayLike, optional
        3-element list or array with the x,y,z location of the root point.
        If None, the most distant tip is set to root.
    root_point_resolution : npt.ArrayLike, optional
        3-element list or array with the x,y,z resolution of the root point, in nanometers per voxel dimension.
    collapse_soma : bool, optional
        If True, collapse nearby vertices into the root point.
    collapse_radius : int, optional
        Distance (in nanometers) for soma collapse.
    nan_rounds : int, optional
        If vertices are missing (or not computed), this sets the number of iterations for smoothing over them.
    require_complete : bool, optional
        If True, raise an Exception if any vertices are missing from the cache.
    level2_graph : np.ndarray, optional
        Level 2 graph for the root id as returned by client.chunkedgraph.level2_chunk_graph.
        A list of lists of edges between level 2 chunks, as defined by their chunk ids.
        If None, will query the chunkedgraph for the level 2 graph. Optional, by default None.

    Returns
    -------
    sk : meshparty.skeleton.Skeleton
        Skeleton for the root id
    mesh : meshparty.trimesh_io.Mesh, optional
        Mesh graph that the skeleton is based on, only returned if return_mesh is True.
    (l2dict_skel, l2dict_skel_reverse): (dict, dict), optional
        Dictionaries mapping l2 ids to skeleton vertices and skeleton vertices to l2 ids, respectively. Only returned if return_l2dict is True.
    (l2dict_mesh, l2dict_mesh_reverse): (dict, dict), optional
        Dictionaries mapping l2 ids to mesh graph vertices and mesh_graph vertices to l2 ids, respectively. Only returned if return_l2dict is True.
    """
    if client is None:
        client = CAVEclient(datastack_name)
 
    if root_point_resolution is None and root_point is not None:
        if cv is None:
            cv = client.info.segmentation_cloudvolume(progress=False)
        if cv is None:
            raise ValueError(
                "Must provide either a client or cv object to get the root_point_resolution."
            )
        root_point_resolution = cv.mip_resolution(0)
    if root_point is not None:
        root_point = np.array(root_point) * root_point_resolution

    mesh, l2dict_mesh, l2dict_r_mesh = pcg_graph(
        root_id,
        client=client,
        return_l2dict=True,
        nan_rounds=nan_rounds,
        require_complete=require_complete,
        level2_graph=level2_graph,
    )

    metameta = {"space": "l2cache", "datastack": client.datastack_name}
    if len(mesh.vertices) == 1:
        vertices = np.array([mesh.vertices[0], mesh.vertices[0]])
        sk = skeletonize.Skeleton(
            vertices=vertices,
            edges=np.array([0, 1]).reshape((1, 2)),
            root=0,
            mesh_to_skel_map=np.array([0]),
            mesh_index=np.array([0, 0]),
            remove_zero_length_edges=False,
            meta={
                "root_id": root_id,
                "skeleton_type": skeleton_type,
                "meta": metameta,
            },
        )
        l2dict, l2dict_r = l2dict_mesh.copy(), l2dict_r_mesh.copy()
        # Assign a fake l2id of zero to the duplicate vertex
        l2dict_r[1] = 0
        l2dict[0] = 1

    else:
        sk = skeletonize.skeletonize_mesh(
            mesh,
            invalidation_d=invalidation_d,
            soma_pt=root_point,
            collapse_soma=collapse_soma,
            soma_radius=collapse_radius,
            compute_radius=False,
            cc_vertex_thresh=0,
            remove_zero_length_edges=True,
            meta={
                "root_id": root_id,
                "skeleton_type": skeleton_type,
                "meta": metameta,
            },
        )
        l2dict, l2dict_r = sk_utils.filter_l2dict(sk, l2dict_r_mesh)

    out_list = [sk]
    if return_mesh:
        out_list.append(mesh)
    if return_l2dict:
        out_list.append((l2dict, l2dict_r))
    if return_l2dict_mesh:
        out_list.append((l2dict_mesh, l2dict_r_mesh))
    if len(out_list) == 1:
        return out_list[0]
    else:
        return tuple(out_list)


def pcg_meshwork(
    root_id: int,
    datastack_name: Optional[str] = None,
    client: Optional[CAVEclient] = None,
    cv: Optional[cloudvolume.CloudVolume] = None,
    root_point: Optional[list] = None,
    root_point_resolution: Optional[list] = None,
    collapse_soma: bool = False,
    collapse_radius: Numeric = DEFAULT_COLLAPSE_RADIUS,
    synapses: Optional[Union[bool, str]] = None,
    synapse_table: Optional[str] = None,
    remove_self_synapse: bool = True,
    synapse_reference_tables: dict = {},
    live_query: bool = False,
    timestamp: Optional[datetime.datetime] = None,
    invalidation_d: Numeric = DEFAULT_INVALIDATION_D,
    require_complete: bool = False,
    metadata: bool = False,
    synapse_partners: bool = False,
    synapse_point_resolution: list = [1, 1, 1],
    synapse_representative_point_pre: str = "ctr_pt_position",
    synapse_representative_point_post: str = "ctr_pt_position",
    level2_graph: Optional[np.ndarray] = None,
) -> meshwork.Meshwork:
    """Generate a meshwork file based on the level 2 graph.

    Parameters
    ----------
    root_id : int
        Root id of an object in the pychunkedgraph.
    datastack_name : str or None, optional
        Datastack name to use to initialize a client, if none is provided. By default None.
    client : caveclient.CAVEclientFull or None, optional
        Initialized CAVE client. If None is given, will use the datastack_name to create one. By default None
    cv : cloudvolume.CloudVolume or None, optional
        Initialized cloudvolume. If none is given, the client info will be used to create one. By default None
    root_point : array-like or None, optional
        3 element xyz location for the location to set the root in units set by root_point_resolution,
        by default None. If None, a distal tip is selected.
    root_point_resolution : array-like, optional
        Resolution in euclidean space of the root_point, by default [4, 4, 40]
    collapse_soma : bool, optional,
        If True, collapses vertices within a given radius of the root point into the root vertex, typically to better
        represent primary neurite branches. Requires a specified root_point. Default if False.
    collapse_radius : float, optional
        Max distance in euclidean space for soma collapse. Default is 10,000 nm (10 microns).
    synapses : 'pre', 'post', 'all', True, or None, optional
        If not None, queries the synapse_table for presynaptic synapses (if 'pre'),  postsynaptic sites (if 'post'), or both (if 'all' or True). By default None
    synapse_table : str, optional
        Name of the synapse table to query if synapses are requested, by default None
    remove_self_synapse : bool, optional
        If True, filters out synapses whose pre- and postsynaptic root ids are the same neuron, by default True
    synapse_reference_tables : dict, optional
        A dictionary of synapse reference tables to attach to synapses.
        Keys are 'pre' and 'post', values are table names. Defaults to {}.
    live_query : bool, optional
        If True, expect a timestamp for querying at a give point in time. Otherwise, use the materializatio set by the client. Optional, by default False.
    timestamp = datetime.datetime, optional
        If set, acts as the time at which all root ids and annotations are found at.
    invalidation_d : int, optional
        Invalidation radius in hops for the mesh skeletonization along the chunk adjacency graph, by default 3
    require_complete : bool, optional
        If True, raise an Exception if any vertices are missing from the cache, by default False
    metadata : bool, optional
        If True, adds metadata to the meshwork annotations. By default False.
    synapse_partners : bool, optional
        If True, includes the partner root id to the synapse annotation. By default False, because partner roots can change across time.
    synapse_point_resolution : array-like, optional
        Resolution in euclidean space of the synapse points, by default None. If None, the resolution will be the default of the synapse table.
    synapse_representative_point_pre : str, optional
        If set, uses the specified column in the synapse table for the pre-synaptic points. By default 'ctr_pt_position'.
    synapse_representative_point_post : str, optional
        If set, uses the specified column in the synapse table for the post-synaptic points. By default 'ctr_pt_position'.

    Returns
    -------
    meshparty.meshwork.Meshwork
        Meshwork object with skeleton based on the level 2 graph. See documentation for details.
    """
    if client is None:
        client = CAVEclient(datastack_name)
    if root_point_resolution is None and root_point is not None:
        if cv is None:
            cv = client.info.segmentation_cloudvolume(progress=False, parallel=1)
        if cv is None:
            raise ValueError(
                "Must provide either a client or cv object to get the root_point_resolution."
            )
        root_point_resolution = cv.mip_resolution(0)
    if synapse_table is None and synapses is not None:
        synapse_table = client.materialize.synapse_table

    sk, mesh, (l2dict_mesh, l2dict_mesh_r) = pcg_skeleton(
        root_id,
        client=client,
        cv=cv,
        root_point=root_point,
        root_point_resolution=root_point_resolution,
        collapse_soma=collapse_soma,
        collapse_radius=collapse_radius,
        invalidation_d=invalidation_d,
        return_mesh=True,
        return_l2dict_mesh=True,
        require_complete=require_complete,
        level2_graph=level2_graph,
    )

    nrn = meshwork.Meshwork(mesh, seg_id=root_id, skeleton=sk)

    pre, post = False, False
    if synapses is not None and synapse_table is not None:
        if synapses == "pre":
            pre, post = True, False
        elif synapses == "post":
            pre, post = False, True
        elif synapses == "all" or synapses is True:
            pre, post = True, True
        else:
            raise ValueError('Synapses must be one of "pre", "post", or "all" or True.')

        if not timestamp:
            timestamp = client.materialize.get_timestamp()

        features.add_synapses(
            nrn,
            synapse_table,
            l2dict_mesh,
            client,
            root_id=root_id,
            pre=pre,
            post=post,
            remove_self_synapse=remove_self_synapse,
            timestamp=timestamp,
            live_query=live_query,
            metadata=metadata,
            synapse_partners=synapse_partners,
            synapse_point_resolution=synapse_point_resolution,
            synapse_representative_point_pre=synapse_representative_point_pre,
            synapse_representative_point_post=synapse_representative_point_post,
            synapse_reference_tables=synapse_reference_tables,
        )

    features.add_lvl2_ids(nrn, l2dict_mesh)
    return nrn


def coord_space_skeleton(
    root_id,
    client,
    datastack_name=None,
    cv=None,
    invalidation_d=10_000,
    return_mesh=False,
    return_l2dict=False,
    return_l2dict_mesh=False,
    root_point=None,
    root_point_resolution=None,
    collapse_soma=False,
    collapse_radius=7500,
    nan_rounds=10,
    require_complete=False,
):
    """Produce a skeleton from the level 2 graph.

    **Deprecated: Please use pcg_skeleton instead.**

    Parameters
    ----------
    root_id : int
        Root id of a segment
    client : CAVEclient.caveclient
        Initialized CAVEclient for the dataset.
    datastack_name : string, optional
        If client is None, initializes a CAVEclient at this datastack, by default None.
    cv : cloudvolume.CloudVolume, optional
        Initialized CloudVolume object for the dataset. This does not replace the caveclient, but
        a pre-initizialized cloudvolume can save some time during batch processing. By default None.
    invalidation_d : int, optional
        Distance (in nanometers) for TEASAR skeleton invalidation, by default 10_000.
    return_mesh : bool, optional
        If True, returns the mesh graph as well as the skeleton, by default False
    return_l2dict : bool, optional
        If True, returns the mappings between l2 ids and skeleton vertices, by default False
    return_l2dict_mesh : bool, optional
        If True, returns mappings between l2 ids and mesh graph vertices, by default False
    root_point : list-like, optional
        3-element list or array with the x,y,z location of the root point. Optional, by default None.
        If None, the most distant tip is set to root.
    root_point_resolution : list-like, optional
        3-element list or array with the x,y,z resolution of the root point, in nanometers per voxel dimension, by default None.
    collapse_soma : bool, optional
        If True, collapse nearby vertices into the root point, by default False.
    collapse_radius : int, optional
        Distance (in nanometers) for soma collapse, by default 7500.
    nan_rounds : int, optional
        If vertices are missing (or not computed), this sets the number of iterations for smoothing over them. By default 10
    require_complete : bool, optional
        If True, raise an Exception if any vertices are missing from the cache, by default False

    Returns
    -------
    sk : meshparty.skeleton.Skeleton
        Skeleton for the root id
    mesh : meshparty.trimesh_io.Mesh (optional)
        Mesh graph that the skeleton is based on, only returned if return_mesh is True.
    (l2dict_skel, l2dict_reverse): tuple of dicts (optional)
        Dictionaries mapping l2 ids to skeleton vertices and skeleton vertices to l2 ids, respectively. Only returned if return_l2dict is True.
    (l2dict_mesh, l2dict_mesh): tuple of dicts (optional)
        Dictionaries mapping l2 ids to mesh graph vertices and mesh_graph vertices to l2 ids, respectively. Only returned if return_l2dict is True.
    """
    warnings.warn(
        "The function `coord_space_skeleton` is deprecated and has been replaced with 'pcg_skeleton'",
        DeprecationWarning,
    )
    return pcg_skeleton(
        root_id,
        client,
        datastack_name=datastack_name,
        cv=cv,
        invalidation_d=invalidation_d,
        return_mesh=return_mesh,
        return_l2dict=return_l2dict,
        return_l2dict_mesh=return_l2dict_mesh,
        root_point=root_point,
        root_point_resolution=root_point_resolution,
        collapse_soma=collapse_soma,
        collapse_radius=collapse_radius,
        nan_rounds=nan_rounds,
        require_complete=require_complete,
    )


def coord_space_mesh(
    root_id,
    client,
    cv=None,
    return_l2dict=False,
    nan_rounds=10,
    require_complete=False,
):
    """Compute the level 2 spatial graph (or mesh) of a given root id using the l2cache.
    Deprecated: Use pcg_graph instead.

    Parameters
    ----------
    root_id : int
        Root id of a segment
    client : CAVEclient.caveclient
        Initialized CAVEclient for the dataset.
    cv : cloudvolume.CloudVolume, optional
        Initialized CloudVolume object for the dataset. This does not replace the caveclient, but
        a pre-initizialized cloudvolume can save some time during batch processing. By default None.
    return_l2dict : bool, optional
        If True, returns the mappings between l2 ids and vertices, by default False
    nan_rounds : int, optional
        If vertices are missing (or not computed), this sets the number of iterations for smoothing over them. By default 10
    require_complete : bool, optional
        If True, raise an Exception if any vertices are missing from the cache, by default False

    Returns
    -------
    mesh : meshparty.trimesh_io.Mesh
        Object with a vertex for every level 2 id and edges between all connected level 2 chunks.
    l2dict : dict (optional)
        Dictionary with keys as level 2 ids and values as mesh vertex index. Optional, only returned if `return_l2dict` is True.
    l2dict_reverse : dict (optional)
        Dictionary with keys as mesh vertex indices and values as level 2 id. Optional, only returned if `return_l2dict` is True.
    """

    warnings.warn(
        "The function `coord_space_mesh` is deprecated and has been replaced with 'pcg_graph'",
        DeprecationWarning,
    )
    return pcg_graph(
        root_id,
        client,
        cv=cv,
        return_l2dict=return_l2dict,
        nan_rounds=nan_rounds,
        require_complete=require_complete,
    )


def coord_space_meshwork(
    root_id,
    datastack_name=None,
    client=None,
    cv=None,
    root_point=None,
    root_point_resolution=None,
    collapse_soma=False,
    collapse_radius=DEFAULT_COLLAPSE_RADIUS,
    synapses=None,
    synapse_table=None,
    remove_self_synapse=True,
    live_query=False,
    timestamp=None,
    invalidation_d=DEFAULT_INVALIDATION_D,
    require_complete=False,
    metadata=False,
):
    """Generate a meshwork file based on the level 2 graph.

    Parameters
    ----------
    root_id : int
        Root id of an object in the pychunkedgraph.
    datastack_name : str or None, optional
        Datastack name to use to initialize a client, if none is provided. By default None.
    client : caveclient.CAVEclientFull or None, optional
        Initialized CAVE client. If None is given, will use the datastack_name to create one. By default None
    cv : cloudvolume.CloudVolume or None, optional
        Initialized cloudvolume. If none is given, the client info will be used to create one. By default None
    root_point : array-like or None, optional
        3 element xyz location for the location to set the root in units set by root_point_resolution,
        by default None. If None, a distal tip is selected.
    root_point_resolution : array-like, optional
        Resolution in euclidean space of the root_point, by default [4, 4, 40]
    collapse_soma : bool, optional,
        If True, collapses vertices within a given radius of the root point into the root vertex, typically to better
        represent primary neurite branches. Requires a specified root_point. Default if False.
    collapse_radius : float, optional
        Max distance in euclidean space for soma collapse. Default is 10,000 nm (10 microns).
    synapses : 'pre', 'post', 'all', or None, optional
        If not None, queries the synapse_table for presynaptic synapses (if 'pre'),  postsynaptic sites (if 'post'), or both (if 'all'). By default None
    synapse_table : str, optional
        Name of the synapse table to query if synapses are requested, by default None
    remove_self_synapse : bool, optional
        If True, filters out synapses whose pre- and postsynaptic root ids are the same neuron, by default True
    live_query : bool, optional
        If True, expect a timestamp for querying at a give point in time. Otherwise, use the materializatio set by the client. Optional, by default False.
    timestamp = datetime.datetime, optional
        If set, acts as the time at which all root ids and annotations are found at.
    invalidation_d : int, optional
        Invalidation radius in hops for the mesh skeletonization along the chunk adjacency graph, by default 3
    require_complete : bool, optional
        If True, raise an Exception if any vertices are missing from the cache, by default False
    level2_graph : np.ndarray, optional
        Level 2 graph for the root id as returned by client.chunkedgraph.level2_chunk_graph.
        A list of lists of edges between level 2 chunks, as defined by their chunk ids.
        If None, will query the chunkedgraph for the level 2 graph. Optional, by default None.


    Returns
    -------
    meshparty.meshwork.Meshwork
        Meshwork object with skeleton based on the level 2 graph. See documentation for details.
    """
    warnings.warn(
        "The function `coord_space_meshwork` is deprecated and has been replaced with 'pcg_meshwork'",
        DeprecationWarning,
    )

    return pcg_meshwork(
        root_id,
        datastack_name=datastack_name,
        client=client,
        cv=cv,
        root_point=root_point,
        root_point_resolution=root_point_resolution,
        collapse_soma=collapse_soma,
        collapse_radius=collapse_radius,
        synapses=synapses,
        synapse_table=synapse_table,
        remove_self_synapse=remove_self_synapse,
        live_query=live_query,
        timestamp=timestamp,
        invalidation_d=invalidation_d,
        require_complete=require_complete,
        metadata=metadata,
    )



import numpy as np
from caveclient import CAVEclient
from meshparty import skeleton, skeletonize, trimesh_io, meshwork

from . import chunk_tools, features
from . import skel_utils as sk_utils
from . import utils

DEFAULT_VOXEL_RESOLUTION = [4, 4, 40]
DEFAULT_COLLAPSE_RADIUS = 7500.0
DEFAULT_INVALIDATION_D = 7500

skeleton_type = "pcg_skel"


def chunk_index_mesh(
    root_id,
    client=None,
    datastack_name=None,
    cv=None,
    return_l2dict=False,
):
    """Download a mesh with chunk index vertices

    Parameters
    ----------
    root_id : int
        Root id to download.
    client : CAVEclient, optional
        Preset CAVEclient, by default None.
    datastack_name : str or None, optional
        Datastack to use to initialize a CAVEclient, by default None.
    cv : cloudvolume.CloudVolume or None, optional
        Cloudvolume instance, by default None.
    return_l2dict : bool, optional
        If True, returns both a l2id to vertex dict and the reverse, by default False.

    Returns
    -------
    mesh : trimesh_io.Mesh
        Chunk graph represented as a mesh, with vertices at chunk index locations and edges in the link_edges attribute.
    l2dict_mesh : dict
        l2 id to mesh vertex index dictionary. Only returned if return_l2dict is True.
    l2dict_r_mesh : dict
        Mesh vertex index to l2 id dictionary. Only returned if return_l2dict is True.
    """

    if client is None:
        client = CAVEclient(datastack_name)
    if cv is None:
        cv = client.info.segmentation_cloudvolume(progress=False)

    lvl2_eg = client.chunkedgraph.level2_chunk_graph(root_id)
    eg, l2dict_mesh, l2dict_r_mesh, x_ch = chunk_tools.build_spatial_graph(lvl2_eg, cv)
    mesh_chunk = trimesh_io.Mesh(
        vertices=x_ch,
        faces=[[0, 0, 0]],  # Some functions fail if no faces are set.
        link_edges=eg,
    )
    if return_l2dict:
        return mesh_chunk, l2dict_mesh, l2dict_r_mesh
    else:
        return mesh_chunk


def chunk_index_skeleton(
    root_id,
    client=None,
    datastack_name=None,
    cv=None,
    root_point=None,
    invalidation_d=3,
    return_mesh=False,
    return_l2dict=False,
    return_mesh_l2dict=False,
    root_point_resolution=None,
    root_point_search_radius=300,
    n_parallel=1,
):
    """Generate a basic skeleton with chunked-graph index vertices.

    Parameters
    ----------
    root_id : np.uint64
        Neuron root id
    client : caveclient.CAVEclient, optional
        CAVEclient for a datastack, by default None. If None, you must specify a datastack name.
    datastack_name : str, optional
        Datastack name to create a CAVEclient, by default None. Only used if client is None.
    cv : cloudvolume.CloudVolume, optional
        CloudVolume associated with the object, by default None. If None, one is created based on the client info.
    root_point : array, optional
        Point in voxel space to set the root vertex. By default None, which makes a random tip root.
    invalidation_d : int, optional
        TEASAR invalidation radius in chunk space, by default 3
    return_mesh : bool, optional
        If True, returns the pre-skeletonization mesh with vertices in chunk index space, by default False
    return_l2dict : bool, optional
        If True, returns the level 2 id to vertex index dict. By default True
    n_parallel : int, optional
        Sets number of parallel threads for cloudvolume, by default 1

    Returns
    -------
    sk : meshparty.skeleton.Skeleton
        Skeleton object
    mesh : meshparty.trimesh_io.Mesh
        Mesh object, only if return_mesh is True
    level2_dict : dict
        Level 2 id to vertex map, only if return_l2dict is True.
    """

    if client is None:
        client = CAVEclient(datastack_name)
    if n_parallel is None:
        n_parallel = 1
    if cv is None:
        cv = client.info.segmentation_cloudvolume(progress=True, parallel=1)

    if root_point_resolution is None:
        root_point_resolution = cv.mip_resolution(0)

    mesh_chunk, l2dict_mesh, l2dict_r_mesh = chunk_index_mesh(
        root_id, client=client, cv=cv, return_l2dict=True
    )

    if root_point is not None:
        lvl2_root_chid, lvl2_root_loc = chunk_tools.get_closest_lvl2_chunk(
            root_point,
            root_id,
            client=client,
            cv=None,
            radius=root_point_search_radius,
            voxel_resolution=root_point_resolution,
            return_point=True,
        )  # Need to have cv=None because of a cloudvolume inconsistency
        root_mesh_index = l2dict_mesh[lvl2_root_chid]
    else:
        root_mesh_index = None

    metameta = {"space": "chunk", "datastack": client.datastack_name}
    sk_ch = skeletonize.skeletonize_mesh(
        mesh_chunk,
        invalidation_d=invalidation_d,
        collapse_soma=False,
        compute_radius=False,
        cc_vertex_thresh=0,
        root_index=root_mesh_index,
        remove_zero_length_edges=False,
        meta={
            "root_id": root_id,
            "skeleton_type": skeleton_type,
            "meta": metameta,
        },
    )

    l2dict, l2dict_r = sk_utils.filter_l2dict(sk_ch, l2dict_r_mesh)

    out_list = [sk_ch]
    if return_mesh:
        out_list.append(mesh_chunk)
    if return_l2dict:
        out_list.append((l2dict, l2dict_r))
    if return_mesh_l2dict:
        out_list.append((l2dict_mesh, l2dict_r_mesh))
    if len(out_list) == 1:
        return out_list[0]
    else:
        return tuple(out_list)


def refine_chunk_index_skeleton(
    sk_ch,
    l2dict_reversed,
    cv,
    refine_inds="all",
    scale_chunk_index=True,
    root_location=None,
    nan_rounds=20,
    return_missing_ids=False,
    segmentation_fallback=False,
    fallback_mip=2,
    cache=None,
    save_to_cache=False,
    client=None,
    l2cache=False,
):
    """Refine skeletons in chunk index space to Euclidean space.

    Parameters
    ----------
    sk_ch : meshparty.skeleton.Skeleton
        Skeleton in chunk index space
    l2dict_reversed : dict
        Mapping between skeleton vertex index and level 2 id.
    cv : cloudvolume.CloudVolume
        Associated cloudvolume
    refine_inds : str, None or list-like, optional
        Skeleton indices to refine, 'all', or None. If 'all', does all skeleton indices.
        If None, downloads no index but can use other options.
        By default 'all'.
    scale_chunk_index : bool, optional
        If True, maps unrefined chunk index locations to the center of the chunk in
        Euclidean space, by default True
    root_location : list-like, optional
        3-element euclidean space location to which to map the root vertex location, by default None
    nan_rounds : int, optional
        Number of passes to smooth over any missing values by averaging proximate vertex locations.
        Only used if refine_inds is 'all'. Default is 20.
    return_missing_ids : bool, optional
        If True, returns ids of any missing level 2 meshes. Default is False
    segmentation_fallback : bool, optional
        If True, downloads the segmentation at mip level in fallback_mip to get a location. Very slow. Default is False.
    fallback_mip : int, optional
        The mip level used in segmentation fallback. Default is 2.
    cache : str, optional
        If set to 'service', uses the l2cache service if available available. Otherwise, a filename for a sqlite database storing locations associated with level 2 ids. Default is None.
    save_to_cache : bool, optional
        If using a sqlite database, setting this to True will add values to the cache as downloads occur.
    client : CAVEclient, optional
        If using the l2cache service, provides a client that can access it.
    l2cache : bool, optional,
        Set to True if using a l2cache to localize vertices. Same as setting cache to 'service'. Default is False.

    Returns
    -------
    meshparty.skeleton.Skeleton
        Skeleton with remapped vertex locations
    """
    if nan_rounds is None:
        convert_missing = True
    else:
        convert_missing = False

    if l2cache:
        cache = "service"

    refine_out = chunk_tools.refine_vertices(
        sk_ch.vertices,
        l2dict_reversed=l2dict_reversed,
        cv=cv,
        refine_inds=refine_inds,
        scale_chunk_index=scale_chunk_index,
        convert_missing=convert_missing,
        return_missing_ids=return_missing_ids,
        segmentation_fallback=segmentation_fallback,
        fallback_mip=fallback_mip,
        cache=cache,
        save_to_cache=save_to_cache,
        client=client,
    )
    if return_missing_ids:
        new_verts, missing_ids = refine_out
    else:
        new_verts = refine_out

    if root_location is not None:
        new_verts[sk_ch.root] = root_location

    l2_sk = skeleton.Skeleton(
        vertices=new_verts,
        edges=sk_ch.edges,
        root=sk_ch.root,
        remove_zero_length_edges=False,
        mesh_index=sk_ch.mesh_index,
        mesh_to_skel_map=sk_ch.mesh_to_skel_map,
        meta=sk_ch.meta,
    )
    metameta = {
        "space": "euclidean",
    }
    try:
        l2_sk.meta.update_metameta(metameta)
    except:
        pass

    if isinstance(refine_inds, str) and refine_inds == "all":
        sk_utils.fix_nan_verts(l2_sk, num_rounds=nan_rounds)

    if return_missing_ids:
        return l2_sk, missing_ids
    else:
        return l2_sk


def pcg_skeleton(
    root_id,
    client=None,
    datastack_name=None,
    cv=None,
    refine="all",
    root_point=None,
    root_point_resolution=None,
    root_point_search_radius=300,
    collapse_soma=False,
    collapse_radius=10_000.0,
    invalidation_d=3,
    return_mesh=False,
    return_l2dict=False,
    return_l2dict_mesh=False,
    return_missing_ids=False,
    nan_rounds=20,
    segmentation_fallback=False,
    fallback_mip=2,
    cache=None,
    save_to_cache=False,
    n_parallel=1,
    l2cache=False,
):
    """Create a euclidean-space skeleton from the pychunkedgraph

    Parameters
    ----------
    root_id : uint64
        Root id of the neuron to skeletonize
    client : caveclient.CAVEclientFull or None, optional
        Pre-specified cave client for the pcg. If this is not set, datastack_name must be provided. By default None
    datastack_name : str or None, optional
        If no client is specified, a CAVEclient is created with this datastack name, by default None
    cv : cloudvolume.CloudVolume or None, optional
        Prespecified cloudvolume instance. If None, uses the client info to make one, by default None
    refine : 'all', 'ep', 'bp', 'epbp', 'bpep', or None, optional
        Selects how to refine vertex locations by downloading mesh chunks. Unrefined vertices are placed in the
        center of their chunk in euclidean space.
        * 'all' refines all vertex locations. (Default)
        * 'ep' refines end points only
        * 'bp' refines branch points only
        * 'bpep' or 'epbp' refines both branch and end points.
        * None refines no points.
        * 'chunk' Keeps things in chunk index space.
    root_point : array-like or None, optional
        3 element xyz location for the location to set the root in units set by root_point_resolution,
        by default None. If None, a distal tip is selected.
    root_point_resolution : array-like, optional
        Resolution in euclidean space of the root_point, by default [4, 4, 40]
    root_point_search_radius : int, optional
        Distance in euclidean space to look for segmentation when finding the root vertex, by default 300
    collapse_soma : bool, optional,
        If True, collapses vertices within a given radius of the root point into the root vertex, typically to better
        represent primary neurite branches. Requires a specified root_point. Default if False.
    collapse_radius : float, optional
        Max distance in euclidean space for soma collapse. Default is 10,000 nm (10 microns).
    invalidation_d : int, optional
        Invalidation radius in hops for the mesh skeletonization along the chunk adjacency graph, by default 3
    return_mesh : bool, optional
        If True, returns the mesh in chunk index space, by default False
    return_l2dict : bool, optional
        If True, returns the tuple (l2dict, l2dict_r), by default False.
        l2dict maps all neuron level2 ids to skeleton vertices. l2dict_r maps skeleton indices to their direct level 2 id.
    return_l2dict_mesh : bool, optional
        If True, returns the tuple (l2dict_mesh, l2dict_mesh_r), by default False.
        l2dict_mesh maps neuron level 2 ids to mesh vertices, l2dict_r maps mesh indices to level 2 ids.
    return_missing_ids : bool, optional
        If True, returns level 2 ids that were missing in the chunkedgraph, by default False. This can be useful
        for submitting remesh requests in case of errors.
    nan_rounds : int, optional
        Maximum number of rounds of smoothing to eliminate missing vertex locations in the event of a
        missing level 2 mesh, by default 20. This is only used when refine=='all'.
    segmentation_fallback : bool, optional
        If True, uses the segmentation in cases of missing level 2 meshes. This is slower but more robust.
        Default is True.
    cache : str or None, optional
        If set to 'service', uses the online l2cache service (if available). Otherwise, this is the filename of a sqlite database with cached lookups for l2 ids. Optional, default is None.
    n_parallel : int, optional
        Number of parallel downloads passed to cloudvolume, by default 1
    l2cache : bool, optional
        Set to True if using the l2cache. Equivalent to cache='service'. Default is False.

    Returns
    -------
    sk_l2 : meshparty.skeleton.Skeleton
        Skeleton with vertices in euclidean space
    mesh_l2 : meshparty.mesh.Mesh, optional
        Mesh with vertices in chunk index space. Only if return_mesh is True.
    (l2dict, l2dict_r) : (dict, dict), optional
        Mappings between level 2 ids and skeleton indices. Only if return_l2dict is True.
    (l2dict_mesh, l2dict_mesh_r) : (dict, dict), optional
        Mappings between level 2 ids and mesh indices. Only if return_l2dict_mesh is True.
    missing_ids : np.array, optional
        List of level 2 ids with missing mesh fragments. Only if return_missing_ids is True.
    """

    if client is None:
        client = CAVEclient(datastack_name)
    if n_parallel is None:
        n_parallel = 1
    if cv is None:
        cv = client.info.segmentation_cloudvolume(progress=True, parallel=1)

    if root_point_resolution is None:
        root_point_resolution = cv.mip_resolution(0)

    (
        sk_ch,
        mesh_ch,
        (l2dict, l2dict_r),
        (l2dict_mesh, l2dict_mesh_r),
    ) = chunk_index_skeleton(
        root_id,
        client=client,
        datastack_name=datastack_name,
        cv=cv,
        root_point=root_point,
        root_point_resolution=root_point_resolution,
        root_point_search_radius=root_point_search_radius,
        invalidation_d=invalidation_d,
        return_mesh=True,
        return_mesh_l2dict=True,
        return_l2dict=True,
        n_parallel=n_parallel,
    )
    if refine == "all":
        refine_inds = "all"
    elif refine == "bp":
        refine_inds = sk_ch.branch_points_undirected
    elif refine == "ep":
        refine_inds = sk_ch.end_points_undirected
    elif refine == "epbp" or refine == "bpep":
        refine_inds = np.concatenate(
            (sk_ch.end_points_undirected, sk_ch.branch_points_undirected)
        )
    elif refine == "chunk":
        refine_inds = None
    elif refine is None:
        refine_inds = None
    else:
        raise ValueError(
            '"refine" must be one of "all", "bp", "ep", "epbp"/"bpep", "chunk", or None'
        )

    if root_point is not None:
        root_point_euc = root_point * np.array([root_point_resolution])
    else:
        root_point_euc = None

    sk_l2, missing_ids = refine_chunk_index_skeleton(
        sk_ch,
        l2dict_r,
        cv=cv,
        refine_inds=refine_inds,
        scale_chunk_index=True,
        root_location=root_point_euc,
        nan_rounds=nan_rounds,
        return_missing_ids=True,
        segmentation_fallback=segmentation_fallback,
        fallback_mip=fallback_mip,
        cache=cache,
        save_to_cache=save_to_cache,
        client=client,
        l2cache=l2cache,
    )

    if refine == "chunk" or refine is None:
        refinement_method = None
    elif segmentation_fallback:
        refinement_method = "mesh_average_with_seg_fallback"
    else:
        refinement_method = "mesh_average"
    metameta = {
        "refinement": refine,
        "refinement_method": refinement_method,
        "nan_rounds": nan_rounds,
    }
    try:
        sk_l2.meta.update_metameta(metameta)
    except:
        pass

    if collapse_soma and root_point is not None:
        sk_l2 = collapse_pcg_skeleton(
            sk_l2.vertices[sk_l2.root], sk_l2, collapse_radius
        )

    if refine == "chunk":
        sk_l2._rooted._vertices = utils.nm_to_chunk(sk_l2.vertices, cv)
        try:
            sk_l2.meta.update_metameta({"space": "chunk"})
        except:
            pass

    output = [sk_l2]
    if return_mesh:
        output.append(mesh_ch)
    if return_l2dict:
        output.append((sk_utils.propagate_l2dict(sk_l2, l2dict_mesh), l2dict_r))
    if return_l2dict_mesh:
        output.append((l2dict_mesh, l2dict_mesh_r))
    if return_missing_ids:
        output.append(missing_ids)
    if len(output) == 1:
        return output[0]
    else:
        return tuple(output)


def pcg_meshwork(
    root_id,
    datastack_name=None,
    client=None,
    cv=None,
    refine="all",
    root_point=None,
    root_point_resolution=None,
    root_point_search_radius=300,
    collapse_soma=False,
    collapse_radius=DEFAULT_COLLAPSE_RADIUS,
    synapses=None,
    synapse_table=None,
    remove_self_synapse=True,
    live_query=False,
    timestamp=None,
    invalidation_d=3,
    segmentation_fallback=False,
    fallback_mip=2,
    cache=None,
    save_to_cache=False,
    n_parallel=None,
    l2cache=False,
):
    """Generate a meshwork file based on the level 2 graph.

    Parameters
    ----------
    root_id : int
        Root id of an object in the pychunkedgraph.
    datastack_name : str or None, optional
        Datastack name to use to initialize a client, if none is provided. By default None.
    client : caveclient.CAVEclientFull or None, optional
        Initialized CAVE client. If None is given, will use the datastack_name to create one. By default None
    cv : cloudvolume.CloudVolume or None, optional
        Initialized cloudvolume. If none is given, the client info will be used to create one. By default None
    refine : 'all', 'ep', 'bp', 'epbp'/'bpep', or None, optional
        Selects how to refine vertex locations by downloading mesh chunks.
        Unrefined vertices are placed in the center of their chunk in euclidean space.
        * 'all' refines all vertex locations. (Default)
        * 'ep' refines end points only
        * 'bp' refines branch points only
        * 'bpep' or 'epbp' refines both branch and end points.
        * 'chunk' keeps vertices in chunk index space.
        * None refines no points but maps them to the center of the chunk in euclidean space.
    root_point : array-like or None, optional
        3 element xyz location for the location to set the root in units set by root_point_resolution,
        by default None. If None, a distal tip is selected.
    root_point_resolution : array-like, optional
        Resolution in euclidean space of the root_point, by default [4, 4, 40]
    root_point_search_radius : int, optional
        Distance in euclidean space to look for segmentation when finding the root vertex, by default 300
    collapse_soma : bool, optional,
        If True, collapses vertices within a given radius of the root point into the root vertex, typically to better
        represent primary neurite branches. Requires a specified root_point. Default if False.
    collapse_radius : float, optional
        Max distance in euclidean space for soma collapse. Default is 10,000 nm (10 microns).
    synapses : 'pre', 'post', 'all', or None, optional
        If not None, queries the synapse_table for presynaptic synapses (if 'pre'),  postsynaptic sites (if 'post'), or both (if 'all'). By default None
    synapse_table : str, optional
        Name of the synapse table to query if synapses are requested, by default None
    remove_self_synapse : bool, optional
        If True, filters out synapses whose pre- and postsynaptic root ids are the same neuron, by default True
    invalidation_d : int, optional
        Invalidation radius in hops for the mesh skeletonization along the chunk adjacency graph, by default 3
    cache : str or None, optional
        If set to 'service', uses the online l2cache service (if available). Otherwise, this is the filename of a sqlite database with cached lookups for l2 ids. Optional, default is None.
    n_parallel : int, optional
        Number of parallel downloads passed to cloudvolume, by default 1
    l2cache : bool, optional
        Set to True to use the l2cache. Equivalent to cache='service'.

    Returns
    -------
    meshparty.meshwork.Meshwork
        Meshwork object with skeleton based on the level 2 graph. See documentation for details.
    """

    if client is None:
        client = CAVEclient(datastack_name)
    if n_parallel is None:
        n_parallel = 1
    if cv is None:
        cv = client.info.segmentation_cloudvolume(progress=True, parallel=1)
    if root_point_resolution is None:
        root_point_resolution = cv.mip_resolution(0)

    sk_l2, mesh_chunk, (l2dict_mesh, l2dict_mesh_r) = pcg_skeleton(
        root_id,
        client=client,
        cv=cv,
        root_point=root_point,
        root_point_resolution=root_point_resolution,
        root_point_search_radius=root_point_search_radius,
        collapse_soma=collapse_soma,
        collapse_radius=collapse_radius,
        refine=refine,
        invalidation_d=invalidation_d,
        n_parallel=n_parallel,
        return_mesh=True,
        return_l2dict_mesh=True,
        segmentation_fallback=segmentation_fallback,
        fallback_mip=fallback_mip,
        cache=cache,
        save_to_cache=save_to_cache,
        l2cache=l2cache,
    )

    nrn = meshwork.Meshwork(mesh_chunk, seg_id=root_id, skeleton=sk_l2)

    if synapses is not None and synapse_table is not None:
        if synapses == "pre":
            pre, post = True, False
        elif synapses == "post":
            pre, post = False, True
        elif synapses == "all":
            pre, post = True, True
        else:
            raise ValueError('Synapses must be one of "pre", "post", or "all".')

        if not timestamp:
            timestamp = client.materialize.get_timestamp()

        features.add_synapses(
            nrn,
            synapse_table,
            l2dict_mesh,
            client,
            root_id=root_id,
            pre=pre,
            post=post,
            remove_self_synapse=remove_self_synapse,
            timestamp=timestamp,
            live_query=live_query,
        )

    features.add_lvl2_ids(nrn, l2dict_mesh)

    if refine != "chunk":
        chunk_tools.adjust_meshwork(nrn, cv)

    return nrn



from functools import reduce

import numpy as np
from trimesh import creation


def chunk_dims(cv):
    """Gets the size of a chunk in euclidean space

    Parameters
    ----------
    cv : cloudvolume.CloudVolume
        Chunkedgraph-targeted cloudvolume object

    Returns
    -------
    np.array
        3-element box dimensions of a chunk in nanometers.
    """
    dims = chunk_to_nm([1, 1, 1], cv) - chunk_to_nm([0, 0, 0], cv)
    return np.squeeze(dims)


def nm_to_chunk(xyz_nm, cv):
    """Map a location in euclidean space to a chunk

    Parameters
    ----------
    xyz_nm : array-like
        Nx3 array of spatial points
    cv : cloudvolume.CloudVolume
        CloudVolume object associated with the chunked space
    voxel_resolution : list, optional
        Voxel resolution, by default [4, 4, 40]

    Returns
    -------
    np.array
        Nx3 array of chunk indices
    """
    voxel_resolution = np.array(cv.mip_resolution(0))
    x_vox = np.atleast_2d(xyz_nm) / np.array(voxel_resolution)
    offset_vox = np.array(cv.mesh.meta.meta.voxel_offset(0))
    return (x_vox - offset_vox) / np.array(cv.mesh.meta.meta.graph_chunk_size)


def chunk_to_nm(xyz_ch, cv):
    """Map a chunk location to Euclidean space

    Parameters
    ----------
    xyz_ch : array-like
        Nx3 array of chunk indices
    cv : cloudvolume.CloudVolume
        CloudVolume object associated with the chunked space
    voxel_resolution : list, optional
        Voxel resolution, by default [4, 4, 40]

    Returns
    -------
    np.array
        Nx3 array of spatial points
    """
    x_vox = np.atleast_2d(xyz_ch) * cv.mesh.meta.meta.graph_chunk_size
    return (x_vox + np.array(cv.mesh.meta.meta.voxel_offset(0))) * cv.mip_resolution(0)


def _tmat(xyz):
    """4x4 transformation matrix for simple translation by xyz vector"""
    T = np.eye(4)
    T[0:3, 3] = np.array(xyz).reshape(1, 3)
    return T


def chunk_box(xyz, chunk_size=[1, 1, 1]):
    """Create a trimesh box of a specified size

    Parameters
    ----------
    xyz : array-like
        3-element array for the lower corner of the box
    chunk_size : array-like, optional
        3-element array giving box dimensions, by default [1, 1, 1]

    Returns
    -------
    trimesh.mesh
    """
    xyz_offset = xyz + np.array(chunk_size) / 2
    return creation.box(chunk_size, _tmat(xyz_offset))


def chunk_mesh(xyz_ch, cv):
    """Get a trimesh mesh of a collection of chunks.

    Parameters
    ----------
    xyz_ch : np.array
        Nx3 array of chunk indices
    cv : cloudvolume.CloudVolume
        Home cloudvolume object for the chunks.

    Returns
    -------
    trimesh.mesh
    """
    verts_ch_nm = chunk_to_nm(xyz_ch, cv)
    dim = chunk_dims(cv)

    boxes = [chunk_box(xyz, dim) for xyz in verts_ch_nm]
    boxes_all = reduce(lambda a, b: a + b, boxes)
    return boxes_all



import numpy as np
import pandas as pd

import caveclient
import datetime
from meshparty import meshwork
from meshparty.meshwork.algorithms import split_axon_by_annotation, strahler_order
from warnings import warn
from typing import Optional

from . import pcg_anno

VOL_PROPERTIES = ["area_nm2", "size_nm3", "mean_dt_nm", "max_dt_nm"]


def add_synapse_count(
    nrn: meshwork.Meshwork,
    anno_name: str = "synapse_count",
    pre_syn: str = "pre_syn",
    post_syn: str = "post_syn",
    pre_syn_index: str = "pre_pt_mesh_ind",
    post_syn_index: str = "post_pt_mesh_ind",
    aggregate_size: bool = True,
):
    """
    Create a synapse count label across mesh vertices from existing synapse annotation.

    This function adds a synapse count annotation to the mesh vertices of a neuron. It aggregates
    synapse data from pre-synaptic and post-synaptic annotations and optionally calculates the mean
    size of the synapses.

    Parameters
    ----------
    nrn : meshwork.Meshwork
        The neuron meshwork object containing the mesh and annotations.
    anno_name : str, optional
        The name of the annotation table to be created, by default "synapse_count".
    pre_syn : str, optional
        The name of the pre-synaptic annotation table, by default "pre_syn".
        If set to None, presynaptic points will be skipped.
    post_syn : str, optional
        The name of the post-synaptic annotation table, by default "post_syn".
        If set to None, postsynaptic points will be skipped
    pre_syn_index : str, optional
        The mesh index column for pre-synaptic annotation, by default "pre_pt_mesh_ind".
    post_syn_index : str, optional
        The mesh index column for post-synaptic annotation, by default "post_pt_mesh_ind".
    aggregate_size : bool, optional
        Whether to aggregate the size of synapses, by default True.

    Returns
    -------
    None
        The function modifies the neuron meshwork object in place by adding the synapse count annotation.
    """

    with nrn.mask_context(np.ones(nrn.mesh.unmasked_size).astype(bool)):
        syn_count_df = pd.DataFrame(index=nrn.mesh_indices)

        if pre_syn:
            syn_out_agg_kwargs = {
                "num_syn_out": pd.NamedAgg("id", "count"),
            }
            if aggregate_size:
                syn_out_agg_kwargs["net_size_out"] = pd.NamedAgg("size", "sum")

            syn_out_df = (
                nrn.anno[pre_syn].df.groupby(pre_syn_index).agg(**syn_out_agg_kwargs)
            )

        if post_syn:
            syn_in_agg_kwargs = {
                "num_syn_in": pd.NamedAgg("id", "count"),
            }
            if aggregate_size:
                syn_in_agg_kwargs["net_size_in"] = pd.NamedAgg("size", "sum")

            syn_in_df = (
                nrn.anno[post_syn].df.groupby(post_syn_index).agg(**syn_in_agg_kwargs)
            )

        syn_count_df = (
            syn_count_df.merge(
                syn_out_df,
                left_index=True,
                right_index=True,
                how="left",
            )
            .merge(
                syn_in_df,
                left_index=True,
                right_index=True,
                how="left",
            )
            .fillna(0)
        )

        if pre_syn:
            syn_count_df["num_syn_out"] = syn_count_df["num_syn_out"].astype(int)
        if post_syn:
            syn_count_df["num_syn_in"] = syn_count_df["num_syn_in"].astype(int)
        syn_count_df = syn_count_df.reset_index().rename(
            columns={"index": "mesh_index"}
        )

        nrn.anno.add_annotations(
            anno_name,
            syn_count_df,
            index_column="mesh_index",
        )
    pass


def add_synapses(
    nrn: meshwork.Meshwork,
    synapse_table: str,
    l2dict_mesh: dict,
    client: caveclient.CAVEclient,
    root_id: int = None,
    pre: bool = False,
    post: bool = False,
    remove_self_synapse: bool = True,
    timestamp: datetime.datetime = None,
    live_query: bool = False,
    metadata: bool = False,
    synapse_partners: bool = False,
    synapse_point_resolution: list[float] = None,
    synapse_representative_point_pre: str = "ctr_pt_position",
    synapse_representative_point_post: str = "ctr_pt_position",
    synapse_reference_tables: dict = {},
) -> None:
    """Add synapses to a meshwork object based on l2ids

    Parameters
    ----------
    nrn : meshparty.meshwork.Meshwork
        Meshwork object
    synapse_table : str
        Annotation table to use for synapses
    l2dict_mesh : dict
        Dictionary mapping l2ids to vertex ids
    client : caveclient.CAVEclient
        Caveclient to use to get annotations
    root_id : int, optional
        Root id of the cell, by default None. If none, uses the root id set in nrn.seg_id.
    pre : bool, optional
        If True, add presynaptic synpases (i.e. outputs), by default False
    post : bool, optional
        If True, add postsynaptic synapses (i.e. inputs), by default False
    remove_self_synapse : bool, optional
        If True, omit synapses where the pre and post root ids are the same, by default True
    timestamp : datetime.datetime, optional
        Datetime to use for root id lookups if not using a materialized version, by default None
    live_query : bool, optional
        If True, use a timestamp to look up root ids, by default False
    synapse_partners : bool, optional
        If True, returns the root id of synapses partners in the dataframe.
        By default, this is False because partner root ids change with editing and are not specified by this cell's data alone.
    """
    if root_id is None:
        root_id = nrn.seg_id

    pre_syn_df, post_syn_df = pcg_anno.get_level2_synapses(
        root_id,
        l2dict_mesh,
        client,
        synapse_table,
        remove_self=remove_self_synapse,
        pre=pre,
        post=post,
        live_query=live_query,
        timestamp=timestamp,
        metadata=metadata,
        synapse_point_resolution=synapse_point_resolution,
        synapse_reference_tables=synapse_reference_tables,
    )

    if pre_syn_df is not None:
        if not synapse_partners:
            pre_syn_df = pre_syn_df.drop(columns=["post_pt_root_id"])
        nrn.anno.add_annotations(
            "pre_syn",
            pre_syn_df,
            index_column="pre_pt_mesh_ind",
            point_column=synapse_representative_point_pre,
            voxel_resolution=pre_syn_df.attrs.get("dataframe_resolution"),
        )
    if post_syn_df is not None:
        if not synapse_partners:
            post_syn_df = post_syn_df.drop(columns=["pre_pt_root_id"])
        nrn.anno.add_annotations(
            "post_syn",
            post_syn_df,
            index_column="post_pt_mesh_ind",
            point_column=synapse_representative_point_post,
            voxel_resolution=post_syn_df.attrs.get("dataframe_resolution"),
        )


def add_lvl2_ids(
    nrn: meshwork.Meshwork,
    l2dict_mesh: dict,
    property_name: str = "lvl2_ids",
) -> None:
    """Add meshwork annotation table associating level 2 ids with vertex ids.

    Parameters
    ----------
    nrn : meshparty.meshwork.Meshwork
        Meshwork object
    l2dict_mesh : dict
        Dictionary mapping L2 ids to mesh graph indices.
    property_name : str, optional
        Name of the annotation table, by default "lvl2_ids"
    """
    lvl2_df = pd.DataFrame(
        {"lvl2_id": list(l2dict_mesh.keys()), "mesh_ind": list(l2dict_mesh.values())}
    )
    nrn.anno.add_annotations(property_name, lvl2_df, index_column="mesh_ind")


def add_volumetric_properties(
    nrn: meshwork.Meshwork,
    client: caveclient.CAVEclient,
    attributes: list[str] = VOL_PROPERTIES,
    l2id_anno_name: str = "lvl2_ids",
    l2id_col_name: str = "lvl2_id",
    property_name: str = "vol_prop",
) -> None:
    """Add L2 Cache properties as an annotation property table.

    Parameters
    ----------
    nrn : meshparty.meshwork.Meshwork
        Meshwork object
    client : caveclient.CAVEclient
        Initialized caveclient
    attributes : list, optional
        List of attributes to download, by default: ["area_nm2", "size_nm3", "mean_dt_nm", "max_dt_nm"].
    l2id_anno_name : str, optional
        Name of the annotation property table holding L2 ids, by default "lvl2_ids"
    l2id_col_name : str, optional
        Name of the column in the property table holding L2 ids, by default "lvl2_id"
    property_name : str, optional
        Name of the new volume property table, by default "vol_prop"
    """
    l2ids = nrn.anno[l2id_anno_name].df[l2id_col_name]
    dat = client.l2cache.get_l2data(l2ids, attributes=attributes)
    dat_df = pd.DataFrame.from_dict(dat, orient="index")
    dat_df.index = [int(x) for x in dat_df.index]

    l2_df = nrn.anno.lvl2_ids.df
    nrn.anno.add_annotations(
        property_name,
        data=l2_df.merge(dat_df, left_on=l2id_col_name, right_index=True).drop(
            columns=[l2id_col_name]
        ),
        index_column="mesh_ind",
    )


def add_segment_properties(
    nrn: meshwork.Meshwork,
    segment_property_name: str = "segment_properties",
    effective_radius: bool = True,
    area_factor: bool = True,
    strahler: bool = True,
    strahler_by_compartment: bool = False,
    volume_property_name: str = "vol_prop",
    volume_col_name: str = "size_nm3",
    area_col_name: str = "area_nm2",
    root_as_sphere: bool = True,
    comp_mask: str = "is_axon",
) -> None:
    """Use volumetric and topological properties to add descriptive properties for each skeleton vertex.
    Note that properties are estimated per segment, the unbranched region between branch points and/or endpoints.

    This function assumes that the volume properties have already been added to the
    meshwork, which can be done using `add_volumetric_properties`.

    Parameters
    ----------
    nrn : meshparty.meshwork.Meshwork
        Meshwork object
    segment_property_name : str, optional
        Name of the new annotation property table, by default "segment_properties"
    effective_radius : bool, optional
        If True, add a column for a radius estimate found by computing the radius of an equivalent cylindar with the same length and volume as each segment, by default True
    area_factor : bool, optional
        If True, add a column with the ratio of the surface area of the segment and the surface area of an equivalent cylinder (without endcaps), by default True
    strahler : bool, optional
        If True, add a column with the strahler number of the segment, by default True
    strahler_by_compartment : bool, optional
        If True, computer strahler number for axon and dendrite separately, using the annotation property table specified in `comp_mask`, by default False
    volume_property_name : str, optional
        Name of the volume properties table as generated by the function `add_volumetric_properties`, by default "vol_prop"
    volume_col_name : str, optional
        Name of the column holding volume, by default "size_nm3"
    area_col_name : str, optional
        Name of the column holding surface area, by default "area_nm2"
    root_as_sphere : bool, optional
        Treats the root location as a sphere for setting the effective radius, by default True
    comp_mask : str, optional
        Sets the annotation table to mask off for strahler number computation, by default "is_axon".
    """
    if volume_property_name not in nrn.anno.table_names:
        msg = f"Volume property table {volume_property_name} not found in `nrn.anno`. "
        msg += "You may want to run `add_volumetric_properties` on this neuron first "
        msg += "before calling this function."
        raise ValueError(msg)

    seg_num = []
    is_root = []
    segment_index = np.zeros(len(nrn.vertices), dtype=int)
    if strahler:
        if strahler_by_compartment:
            with nrn.mask_context(nrn.anno[comp_mask].mesh_mask) as nrnf:
                so_axon = strahler_order(nrnf)
                so_axon_base = np.full(len(nrnf.mesh_mask), np.nan)
                so_axon_base[nrnf.mesh_mask] = so_axon
            with nrn.mask_context(~nrn.anno[comp_mask].mesh_mask) as nrnf:
                so_dend = strahler_order(nrnf)
                so_dend_base = np.full(len(nrnf.mesh_mask), np.nan)
                so_dend_base[nrnf.mesh_mask] = so_dend
            so = np.where(~np.isnan(so_axon_base), so_axon_base, so_dend_base).astype(
                int
            )
        if not strahler_by_compartment:
            so = strahler_order(nrn)
        seg_strahler = []

    if effective_radius:
        seg_vols = []
        seg_pls = []
    if area_factor:
        seg_areas = []
    prop_df = nrn.anno[volume_property_name].df
    prop_df.set_index("mesh_ind")
    for ii, seg in enumerate(nrn.segments()):
        seg_num.append(ii)
        is_root.append(nrn.root in seg)
        segment_index[seg] = ii
        if effective_radius:
            seg_vols.append(prop_df.loc[seg][volume_col_name].sum())
            if nrn.skeleton.root in seg.to_skel_index:
                seg_pls.append(0)
            else:
                seg_plus = nrn.MeshIndex(
                    np.unique(np.concatenate((seg, nrn.parent_index(seg))))
                )  # Add dist to parent node for path length
                seg_pls.append(nrn.path_length(seg_plus))
        if area_factor:
            seg_areas.append(prop_df.loc[seg][area_col_name].sum())
        if strahler:
            seg_strahler.append(so[seg[0]])

    base_df = pd.DataFrame(
        {
            "seg_num": segment_index,
            "mesh_ind": np.arange(len(segment_index)),
        }
    )
    prop_dict = {"seg_num": seg_num, "is_root": is_root}
    if effective_radius:
        prop_dict["vol"] = seg_vols
        prop_dict["len"] = seg_pls
    if area_factor:
        prop_dict["area"] = seg_areas
    if strahler:
        prop_dict["strahler"] = seg_strahler
    prop_df = pd.DataFrame(prop_dict)
    if effective_radius:
        prop_df["r_eff"] = np.sqrt(prop_df["vol"] / (np.pi * prop_df["len"]))
        if root_as_sphere:
            r_idx = prop_df.query("is_root").index
            prop_df.loc[r_idx, "r_eff"] = (
                prop_df.loc[r_idx, "vol"] * (3 / 4) / np.pi
            ) ** (1 / 3)
    if area_factor and effective_radius:
        prop_df["area_factor"] = prop_df["area"] / (
            2 * np.pi * prop_df["r_eff"] * prop_df["len"]
        )
    base_df = base_df.merge(
        prop_df,
        how="left",
        on="seg_num",
    )
    nrn.anno.add_annotations(
        segment_property_name,
        data=base_df,
        index_column="mesh_ind",
    )


def add_is_axon_annotation(
    nrn: meshwork.Meshwork,
    pre_anno: str,
    post_anno: str,
    annotation_name: str = "is_axon",
    threshold_quality: float = 0.5,
    extend_to_segment: bool = True,
    n_times: int = 1,
    return_quality: bool = False,
):
    """Add an annotation property table specifying which vertices belong to the axon, based on synaptic input and output locations
    For the synapse flow centrality algorithm, see "Quantitative neuroanatomy for connectomics in Drosophila", Schneider-Mizell et al. eLife 2016.

    Parameters
    ----------
    nrn : meshparty.meshwork.Meshwork
        Meshwork object
    pre_anno : str
        Annotation property table name for presynaptic sites (outputs).
    post_anno : str
        Annotation property table name for postsyanptic sites (inputs).
    annotation_name : str, optional
        Name of the new table specifying axon indices, by default "is_axon"
    threshold_quality : float, optional
        Value between 0 and 1 setting the lower limit on input/output segregation quality, by default 0.5. If the segregatation quality is lower than this,
        a table is added but no vertices will be in the table.
    extend_to_segment : bool, optional
        If True, the axon split point will be moved to the most root-ward point on the segment containing the highest synapse flow centrality, by default True
    n_times : int, optional
        The number of times to run axon/dendrite detection in a row, by default 1.
        This should be set to the number of distinct axon branches on a cell, which surprisingly can be more than one even for mouse neurons.
    return_quality : bool, optional
        If True, returns the split quality score (a float between 0-1). Default is False.

    Returns
    -------
    split_quality
        Float between 0-1, with higher numbers indicating a better axon/dendrite segregation.
    """
    is_axon, sq = split_axon_by_annotation(
        nrn,
        pre_anno,
        post_anno,
        return_quality=True,
        extend_to_segment=extend_to_segment,
        n_times=n_times,
    )
    if sq < threshold_quality:
        nrn.anno.add_annotations(annotation_name, [], mask=True)
        warn("Split quality below threshold, no axon mesh vertices added!")
    else:
        nrn.anno.add_annotations(annotation_name, is_axon, mask=True)
    if return_quality:
        return sq
    else:
        pass


def l2dict_from_anno(
    nrn: meshwork.Meshwork,
    table_name: str = "lvl2_ids",
    l2id_col: str = "lvl2_id",
    mesh_ind_col: str = "mesh_ind",
) -> dict:
    return nrn.anno[table_name].df.set_index(l2id_col)[mesh_ind_col].to_dict()


def aggregate_property_to_skeleton(
    nrn: meshwork.Meshwork,
    anno_table: str,
    agg_dict: Optional[dict] = None,
    fill_value: float = 0,
) -> pd.DataFrame:
    """Aggregate a meshwork annotation table to skeletons

    Parameters
    ----------
    nrn : meshwork.Meshwork
        Meshwork object with mesh,skeleton, and annotations
    anno_table : str
        Name of the annotation table
    agg_dict : dict
        Dictionary of column names and aggregation functions that can be used by pandas pd.NamedAgg.
    fill_value : float, optional
        Value to use to fill missing or NaN values, by default 0

    Returns
    -------
    pd.DataFrame
        Dataframe with one row per skeleton index and columns aggregating mesh annotation values
        using the aggregation function specified in the aggregation_dict.
    """
    df = nrn.anno[anno_table].df

    skel_index_column = "skel_index"
    while skel_index_column in df.columns:
        skel_index_column = f"{skel_index_column}_temp"
    df[skel_index_column] = nrn.anno[anno_table].mesh_index.to_skel_index_padded

    aggs = {k: pd.NamedAgg(k, v) for k, v in agg_dict.items()}
    skel_df = df.groupby(skel_index_column).agg(**aggs)
    skel_df = skel_df.reindex(np.arange(0, nrn.skeleton.n_vertices)).fillna(fill_value)
    return skel_df
